#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Integration testing of
 - :mod:`mpop.scene`
 - :mod:`mpop.channel`
 - :mod:`mpop.projector`
"""
import random
import unittest

from mock import MagicMock
import sys
sys.modules['pyresample'] = MagicMock()

import mpop.scene

def random_string(length,
                  choices="abcdefghijklmnopqrstuvwxyz"
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
    """Generates a random string with elements from *set* of the specified
    *length*.
    """
    return "".join([random.choice(choices)
                    for dummy_itr in range(length)])

class TestPPCore(unittest.TestCase):
    """Class for testing the core of mpop.
    """

    def setUp(self):
        """Apply patches.
        """

        self.scene = mpop.scene.SatelliteInstrumentScene()

    def test_channel_list_syntax(self):
        """Test syntax for channel list
        """
        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]
        
        class Satscene(mpop.scene.SatelliteInstrumentScene):
            """Adding a channel list.
            """
            instrument_name = random_string(8)
            channel_list = channels

        self.scene = Satscene()
        for i, chn in enumerate(self.scene.channels):
            self.assertTrue(isinstance(chn, mpop.channel.Channel))
            self.assertEquals(chn.name, channels[i][0])
            self.assertEquals(chn.wavelength_range, list(channels[i][1]))
            self.assertEquals(chn.resolution, channels[i][2])

    def test_project(self):
        """Test project
        """
        channels = [["00_7", (0.5, 0.7, 0.9), 2500],
                    ["06_4", (5.7, 6.4, 7.1), 5000],
                    ["11_5", (10.5, 11.5, 12.5), 5000]]
        
        class Satscene(mpop.scene.SatelliteInstrumentScene):
            """Adding a channel list.
            """
            instrument_name = random_string(8)
            channel_list = channels

        area = random_string(8)
        self.scene = Satscene(area=area)
        area2 = random_string(8)

        new_scene = self.scene.project(area2)
        self.assertEquals(new_scene.area_id, area2)

        for chn in new_scene.channels:
            print chn.area
            self.assertEquals(chn.area, area2)
        
        

def suite():
    """The test suite for test_pp_core.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestPPCore))
    
    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Odim format reader.
"""

filename = "/data/temp/Martin.Raspaud/metop02_20120329_0043_28234_satproj_00000_12959_avhrr.h5"


import h5py

def get_lonlats(filename):


    f__ = h5py.File(filename, "r")

    lat_g = f__["where"]["lat"]["what"].attrs["gain"]
    lat_o = f__["where"]["lat"]["what"].attrs["offset"]
    lat = f__["where"]["lat"]["data"][:] * latg + lato

    lon_g = f__["where"]["lon"]["what"].attrs["gain"]
    lon_o = f__["where"]["lon"]["what"].attrs["offset"]
    lon = f__["where"]["lon"]["data"][:] * lon_g + lon_o

    return lon, lat



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2013, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Batch run stuff.
"""
import datetime
import getopt
import logging
import sys

from mpop.channel import NotLoadedError
from mpop.satellites import get_sat_instr_compositer
from mpop.saturn.tasklist import TaskList
from mpop.satellites import GenericFactory


LOG = logging.getLogger("runner")

def usage(scriptname):
    """Print usefull information for running the script.
    """
    print("""
usage: %s [options]

Available options:
    -r
    --rgb
        Output only the pure rgb images from the product list
    -p
    --pge
        Output only the cloudproduct derived images from the product list
    -d <date_str>
    --date <date_str>
        Run on the specified date (eg. 200910081430)
    -a <area_name>
    --area <area_name>
        Run on the specified area (eg. eurol)
    -c <name>
    --composite <name>
        Output the specified composite. Available composites are:
        overview
        natural
        fog
        nightfog
        convection
        airmass
        ir9
        wv_low
        wv_high
        greensnow
        redsnow
        cloudtop
        hr_overview
        PGE02
        PGE02b
        PGE02bj
        PGE02c
        PGE02cj
        PGE02d
        PGE02e
        PGE03
        CtypeHDF
        NordRad
        CtthHDF
    -v
        Be verbose
    --vv
        Be very verbose
    -h
    --help
        Print this message
          """%scriptname)

def parse_options():
    """Parse command line options.
    """

    time_slots = []
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter("[%(levelname)s: %(asctime)s :"
                                           " %(name)s] %(message)s",
                                           '%Y-%m-%d %H:%M:%S'))
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   "hd:va:c:pr",
                                   ["vv", "help", "date=", "pge", "rgb",
                                    "area=", "composite="])
        del args
    except getopt.GetoptError, err:
        print err
        usage(sys.argv[0])
        sys.exit(2)

    mode = set([])
    areas = set([])
    composites = set([])

    for opt, arg in opts:
        if opt == "-v":
            console.setLevel(logging.INFO)
            logging.getLogger('').addHandler(console)
        elif opt == "--vv":
            console.setLevel(logging.DEBUG)
            logging.getLogger('').addHandler(console)
        elif opt in ("-h", "--help"):
            usage(sys.argv[0])
            sys.exit()
        elif opt in ("-d", "--date"):
            time_slots.append(datetime.datetime.strptime(arg, "%Y%m%d%H%M"))
        elif opt in ("-p", "--pge"):
            mode |= set(["pge"])
        elif opt in ("-r", "--rgb"):
            mode |= set(["rgb"])
        elif opt in ("-a", "--area"):
            areas |= set([arg])
        elif opt in ("-c", "--composites"):
            composites |= set([arg])
        else:
            raise ValueError("Option %s not recognized.")

    return time_slots, mode, areas, composites


class SequentialRunner(object):
    """Runs scenes in a sequential order, as opposed to parallelized running.
    """

    def __init__(self, satellite, instrument, tasklist, precompute=False):
        if isinstance(tasklist, str):
            tasklist = TaskList(tasklist)
        self.tasklist = tasklist
        self.data = None
        self.satellite = satellite[0]
        self.number = satellite[1]
        self.variant = satellite[2]
        self.instrument = instrument

        self.klass = get_sat_instr_compositer((self.satellite,
                                               self.number,
                                               self.variant),
                                              self.instrument)
        self.precompute = precompute
        self.running = True

    def stop(self):
        """Stops the runner.
        """
        self.running = False

    def run_from_cmd(self, hook=None):
        """Batch run mpop.
        """
        time_slots, mode, areas, composites = parse_options()

        tasklist = self.tasklist.shape(self.klass, mode, areas, composites)

        for time_slot in time_slots:
            self.data = GenericFactory.create_scene(self.satellite,
                                                    self.number,
                                                    self.instrument,
                                                    time_slot,
                                                    orbit=None,
                                                    variant=self.variant)
            prerequisites = tasklist.get_prerequisites(self.klass)
            self.data.load(prerequisites)
            self.run_from_data(tasklist, hook)


    def run_from_data(self, tasklist=None, radius=None, hook=None):
        """Run on given data.
        """
        if tasklist is None:
            tasklist = self.tasklist
        for area, productlist in tasklist.items():
            prerequisites = tasklist.get_prerequisites(self.klass, area)
            if not self.running:
                LOG.info("Running interrupted")
                return
            local_data = self.data.project(area, prerequisites, self.precompute,
                                           mode="nearest", radius=radius)
            for product, flist in productlist.items():
                fun = getattr(local_data.image, product)
                flist = flist.put_date(local_data.time_slot)
                metadata = {"area": area}
                LOG.info("Orbit = " + str(local_data.orbit))
                if local_data.orbit is not None:
                    metadata["orbit"] = int(local_data.orbit)
                LOG.info("Instrument Name = " + str(local_data.instrument_name))
                if local_data.instrument_name is not None:
                    metadata["instrument_name"] = str(local_data.instrument_name)

                flist = flist.put_metadata(metadata)
                try:
                    LOG.debug("Doing "+product+".")
                    if not self.running:
                        del local_data
                        LOG.info("Running interrupted")
                        return
                    img = fun()
                    img.info["product_name"] = product
                    img.info["instrument_name"] = metadata["instrument_name"]
                    img.info["start_time"] = self.data.info.get("start_time",
                                                                self.data.time_slot)
                    img.info["end_time"] = self.data.info.get("end_time",
                                                              self.data.time_slot)
                    flist.save_object(img, hook)
                    del img
                except (NotLoadedError, KeyError, ValueError), err:
                    LOG.warning("Error in "+product+": "+str(err))
                    LOG.info("Skipping "+product)
            del local_data

    def run_from_local_data(self, tasklist=None, extra_tags=None, hook=None):
        """Run on given local data (already projected).
        """
        if tasklist is None:
            tasklist = self.tasklist
        metadata = {}

        area_name = self.data.area_id or self.data.area_def.area_id
        tasks, dummy = tasklist.split(area_name)
        if area_name not in tasks:
            LOG.debug("Nothing to do for " + area_name)
            return
        for product, flist in tasks[area_name].items():
            try:
                fun = getattr(self.data.image, product)
            except AttributeError:
                LOG.warning("Missing composite function: " + str(product))
                continue
            flist = flist.put_date(self.data.time_slot)

            if self.data.orbit is not None:
                metadata["orbit"] = int(self.data.orbit)
            LOG.info("Instrument Name = " + str(self.data.instrument_name))
            if self.data.instrument_name is not None:
                metadata["instrument_name"] = str(self.data.instrument_name)
            metadata["satellite"] = self.data.fullname
            metadata["area"] = area_name
            flist = flist.put_metadata(metadata)
            try:
                LOG.debug("Doing "+product+".")
                img = fun()
                img.info["product_name"] = product
                img.info["instrument_name"] = metadata["instrument_name"]
                img.info["start_time"] = self.data.info.get("start_time",
                                                            self.data.time_slot)
                img.info["end_time"] = self.data.info.get("end_time",
                                                            self.data.time_slot)
                if extra_tags:
                    img.tags.update(extra_tags)
                flist.save_object(img, hook)
                del img
            except (NotLoadedError, KeyError), err:
                LOG.warning("Error in "+product+": "+str(err))
                LOG.info("Skipping "+product)

if __name__ == "__main__":
    pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Author(s):
# 
#   Panu Lahtinen <panu.lahtinen@fmi.fi
# 
# This file is part of mpop.
# 
# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import mpop.satin

'''Integration testing of
 - :mod:`mpop.satin`
'''

class TestReaders(unittest.TestCase):
    '''Class for testing mpop.satin'''

    def test_lonlat_to_geo_extent(self):
        '''Test conversion of longitudes and latitudes to area extent.'''

        # MSG3 proj4 string from 
        #  xrit.sat.load(..., only_metadata=True).proj4_params
        proj4_str = 'proj=geos lon_0=0.00 lat_0=0.00 ' \
            'a=6378169.00 b=6356583.80 h=35785831.00'

        # MSG3 maximum extent
        max_extent=(-5567248.07, -5570248.48, 
                     5570248.48, 5567248.07)

        # Few area extents in longitudes/latitudes
        area_extents_ll = [[-68.328121068060341, # left longitude
                             18.363816196771392, # down latitude
                             74.770372053870972, # right longitude
                             75.66494585661934], # up latitude
                           # all corners outside Earth's disc
                           [1e30, 1e30, 1e30, 1e30]
                           ]

        # And corresponding correct values in GEO projection
        geo_extents = [[-5010596.02, 1741593.72, 5570248.48, 5567248.07],
                       [-5567248.07, -5570248.48, 5570248.48, 5567248.07]]

        for i in range(len(area_extents_ll)):
            res = mpop.satin.mipp_xrit.lonlat_to_geo_extent(area_extents_ll[i],
                                                            proj4_str,
                                                            max_extent=\
                                                                max_extent)
            for j in range(len(res)):
                self.assertAlmostEqual(res[j], geo_extents[i][j], 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Abhay Devasthale and Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Sajid Pareeth <sajid.pareeth@fmach.it>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Loader for ascat, netcdf format.
   The driver works for netcdf format of ASCAT soil moisture swath data downloaded from 
   here: http://navigator.eumetsat.int/discovery/Start/DirectSearch/DetailResult.do?f%28r0%29=EO:EUM:DAT:METOP:SOMO12
   rename the CONFIG file mpop/mpop/etc/metop.ascat.cfg.template to metop.cfg to read the ASCAT data
"""
import numpy as np
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import os

from netCDF4 import Dataset


def load(satscene):
    """Load ascat data.
    """

    # Read config file content
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    values = {"orbit": satscene.orbit,
              "satname": satscene.satname,
              "number": satscene.number,
              "instrument": satscene.instrument_name,
              "satellite": satscene.fullname,
              "time_slot": satscene.time_slot,
              "time": satscene.time_slot.strftime('%Y%m%d%H%M%S')
              }
    filename = os.path.join(
        conf.get("ascat-level2", "dir"),
        satscene.time_slot.strftime(conf.get("ascat-level2",
                                             "filename",
                                             raw=True)) % values)
    # Load data from netCDF file
    ds = Dataset(filename, 'r')
    for chn_name in satscene.channels_to_load:
        # Read variable corresponding to channel name
        data = np.ma.masked_array(
            ds.variables[chn_name][:], np.isnan(ds.variables[chn_name][:]))
        satscene[chn_name] = data
    lons = ds.variables['longitude'][:]
    lats = ds.variables['latitude'][:]

    # Set scene area as pyresample geometry object
    try:
        from pyresample import geometry
        satscene.area = geometry.SwathDefinition(lons=lons, lats=lats)
    except ImportError:
        # pyresample not available. Set lon and lats directly
        satscene.area = None
        satscene.lat = lats
        satscene.lon = lons



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2013, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Filelist class.
"""
import logging
import os
import shutil
import tempfile

from mpop.utils import ensure_dir


logger = logging.getLogger(__name__)

class FileList(list):
    """List of files.
    """
    def __init__(self, *args):
        list.__init__(self, *args)

    def put_date(self, date):
        """Return an new filelist with the given *date*.
        """
        return FileList([date.strftime(i) for i in self])

    def put_metadata(self, metadata):
        """Fill in the filelist with the given *metadata*.
        """
        return FileList([i % metadata for i in self])

    def _get_by_ext(self):
        """Convert the filelist into a dict with extensions as keys.
        """
        types = {}
        for filename in self:
            file_tuple = os.path.splitext(filename)
            ext = file_tuple[1][:4]
            types[ext] = types.get(ext, FileList()) + FileList([filename])
        return types

    def save_object(self, obj, hook=None):
        """save *obj* to the filelist.
        """
        files_by_ext = self._get_by_ext()
        for extkey in files_by_ext:
            path, filename = os.path.split(files_by_ext[extkey][0])
            origname = filename
            try:
                ensure_dir(files_by_ext[extkey][0])
                handle, tmpfilename = tempfile.mkstemp(extkey,
                                                       "mpop_tmp",
                                                       path)
                os.fsync(handle)
                obj.save(tmpfilename)
                os.fsync(handle)
                os.chmod(tmpfilename, 0644)
                os.fsync(handle)
            except Exception:
                logger.exception("Something went wrong in saving file... "
                              "Dumping trace.")
                logger.warning("Job skipped, going on with the next.")
                continue
            for filename in files_by_ext[extkey][1:]:
                path2, trash = os.path.split(filename)
                del trash

                ensure_dir(filename)
                handle2, tmpfilename2 = tempfile.mkstemp(extkey,
                                                         "mpop_tmp",
                                                         path2)
                os.fsync(handle2)
                try:
                    shutil.copy(tmpfilename, tmpfilename2)
                    os.fsync(handle2)
                    os.close(handle2)
                except (IOError, OSError):
                    logger.exception("Copying file %s to %s failed"
                                  %(tmpfilename,tmpfilename2))
                    logger.info("Retrying...")
                    try:
                        shutil.copy(tmpfilename, tmpfilename2)
                        os.fsync(handle2)
                        os.close(handle2)
                        logger.info("Went OK this time...")
                    except (IOError, OSError):
                        logger.exception("No way...")
                try:
                    os.rename(tmpfilename2, filename)
                    if hook:
                        hook(obj, filename=origname, uri=filename)
                except (IOError, OSError):
                    logger.exception("Renaming file %s to %s failed"
                                %(tmpfilename2,filename))
                    logger.info("Retrying...")
                    try:
                        os.rename(tmpfilename2, filename)
                        if hook:
                            hook(obj, filename=origname, uri=filename)
                    except (IOError, OSError):
                        logger.exception("No way...")
                logger.debug("Done saving "+filename)
                
            os.rename(tmpfilename, files_by_ext[extkey][0])
            os.fsync(handle)
            os.close(handle)
            if hook:
                hook(obj, filename=origname, uri=files_by_ext[extkey][0])
            logger.debug("Done saving "+files_by_ext[extkey][0])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2011, 2014 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Minimal script for geostationary production.


We take the case of HRIT data from meteosat 9, as send through eumetcast.

- Install mipp, mpop, and pyresample
- Don't forget to set up the PPP_CONFIG_DIR variable to point to your
  configuration files.
- Edit the meteosat09.cfg configuration file (a template is provided in case
  you don't have one) with your HRIT directory:

  .. code-block:: ini

   [seviri-level1]
   format = 'xrit/MSG'
   dir='/data/hrit_in'
   filename='H-000-MSG?__-MSG?________-%(channel)s-%(segment)s-%Y%m%d%H%M-__'
   filename_pro='H-000-MSG?__-MSG?________-_________-%(segment)s-%Y%m%d%H%M-__'
   filename_epi='H-000-MSG?__-MSG?________-_________-%(segment)s-%Y%m%d%H%M-__'

  where `/data/hrit_in` has to be changed to anything that suits your
  environment.

- Here is an example of a minimal script that has to be called as soon as an
  MSG slot has arrived (usually, watching the arrival of the epilogue file
  suffices)

"""

from mpop.utils import debug_on
debug_on()
from mpop.satellites import GeostationaryFactory

import sys
from datetime import datetime

if sys.version_info < (2, 5):
    import time

    def strptime(string, fmt=None):
        """This function is available in the datetime module only
        from Python >= 2.5.
        """

        return datetime(*time.strptime(string, fmt)[:6])

else:
    strptime = datetime.strptime

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print "Usage: " + sys.argv[0] + " time_string"
        sys.exit()

    time_string = sys.argv[1]
    time_slot = strptime(time_string, "%Y%m%d%H%M")
    global_data = GeostationaryFactory.create_scene("meteosat", "09",
                                                    "seviri", time_slot)

    global_data.load()

    areas = ["euro4", "scan2"]

    for area in areas:

        local_data = global_data.project(area)

        img = local_data.image.overview()
        img.save("overview_" + area + "_" + time_string + ".png")

        img = local_data.image.fog()
        img.save("fog_" + area + "_" + time_string + ".png")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Spherical geometry module.
"""

import math
import numpy as np

EPSILON = 0.0000001
            
class Coordinate(object):
    """Point on earth in terms of lat and lon.
    """
    lat = None
    lon = None
    x__ = None
    y__ = None
    z__ = None
    
    def __init__(self, lat=None, lon=None,
                 x__=None, y__=None, z__=None, R=1):

        self.R = R

        if lat is not None and lon is not None:
            self.lat = math.radians(lat)
            self.lon = math.radians(lon)
            self._update_cart()
        else:
            self.x__ = x__
            self.y__ = y__
            self.z__ = z__
            self._update_lonlat()

    def _update_cart(self):
        """Convert lon/lat to cartesian coordinates.
        """

        self.x__ = math.cos(self.lat) * math.cos(self.lon)
        self.y__ = math.cos(self.lat) * math.sin(self.lon)
        self.z__ = math.sin(self.lat)
        

    def _update_lonlat(self):
        """Convert cartesian to lon/lat.
        """
        
        self.lat = math.degrees(math.asin(self.z__ / self.R))
        self.lon = math.degrees(math.atan2(self.y__, self.x__))
        
    def __ne__(self, other):
        if(abs(self.lat - other.lat) < EPSILON and
           abs(self.lon - other.lon) < EPSILON):
            return 0
        else:
            return 1

    def __eq__(self, other):
        return not self.__ne__(other)

    def __str__(self):
        return str((math.degrees(self.lat), math.degrees(self.lon)))
    
    def __repr__(self):
        return str((math.degrees(self.lat), math.degrees(self.lon)))

    def cross2cart(self, point):
        """Compute the cross product, and convert to cartesian coordinates
        (assuming radius 1).
        """

        lat1 = self.lat
        lon1 = self.lon
        lat2 = point.lat
        lon2 = point.lon

        res = Coordinate(
            x__=(math.sin(lat1 - lat2) * math.sin((lon1 + lon2) / 2) *
                 math.cos((lon1 - lon2) / 2) - math.sin(lat1 + lat2) *
                 math.cos((lon1 + lon2) / 2) * math.sin((lon1 - lon2) / 2)),
            y__=(math.sin(lat1 - lat2) * math.cos((lon1 + lon2) / 2) *
                 math.cos((lon1 - lon2) / 2) + math.sin(lat1 + lat2) *
                 math.sin((lon1 + lon2) / 2) * math.sin((lon1 - lon2) / 2)),
            z__=(math.cos(lat1) * math.cos(lat2) * math.sin(lon1 - lon2)))

        return res

    def distance(self, point):
        """Vincenty formula.
        """
        dlambda = self.lon - point.lon
        num = ((math.cos(point.lat) * math.sin(dlambda)) ** 2 +
               (math.cos(self.lat) * math.sin(point.lat) -
                math.sin(self.lat) * math.cos(point.lat) *
                math.cos(dlambda)) ** 2)
        den = (math.sin(self.lat) * math.sin(point.lat) +
               math.cos(self.lat) * math.cos(point.lat) * math.cos(dlambda))

        return math.atan2(math.sqrt(num), den)

    def norm(self):
        """Return the norm of the vector.
        """
        return math.sqrt(self.x__ ** 2 + self.y__ ** 2 + self.z__ ** 2)

    def normalize(self):
        """normalize the vector.
        """

        norm = self.norm()
        self.x__ /= norm
        self.y__ /= norm
        self.z__ /= norm

        return self

    def cross(self, point):
        """cross product with another vector.
        """
        x__ = self.y__ * point.z__ - self.z__ * point.y__
        y__ = self.z__ * point.x__ - self.x__ * point.z__
        z__ = self.x__ * point.y__ - self.y__ * point.x__
        
        return Coordinate(x__=x__, y__=y__, z__=z__)

    def dot(self, point):
        """dot product with another vector.
        """
        return (self.x__ * point.x__ +
                self.y__ * point.y__ +
                self.z__ * point.z__)

class Arc(object):
    """An arc of the great circle between two points.
    """
    start = None
    end = None

    def __init__(self, start, end):
        self.start, self.end = start, end

    def center_angle(self):
        """Angle of an arc at the center of the sphere.
        """
        val = (math.cos(self.start.lat - self.end.lat) +
               math.cos(self.start.lon - self.end.lon) - 1)

        if val > 1:
            val = 1
        elif val < -1:
            val = -1
        
        return math.acos(val)
                           
    def __eq__(self, other):
        if(self.start == other.start and self.end == other.end):
            return 1
        return 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return str((str(self.start), str(self.end)))

    def angle(self, other_arc):
        """Oriented angle between two arcs.
        """
        if self.start == other_arc.start:
            a__ = self.start
            b__ = self.end
            c__ = other_arc.end
        elif self.start == other_arc.end:
            a__ = self.start
            b__ = self.end
            c__ = other_arc.start
        elif self.end == other_arc.end:
            a__ = self.end
            b__ = self.start
            c__ = other_arc.start
        elif self.end == other_arc.start:
            a__ = self.end
            b__ = self.start
            c__ = other_arc.end
        else:
            raise ValueError("No common point in angle computation.")

        ua_ = a__.cross(b__)
        ub_ = a__.cross(c__)

        val =  ua_.dot(ub_) / (ua_.norm() * ub_.norm())
        if abs(val - 1) < EPSILON:
            angle = 0
        elif abs(val + 1) < EPSILON:
            angle = math.pi
        else:
            angle = math.acos(val)    

        n__ = ua_.normalize()
        if n__.dot(c__) > 0:
            return -angle
        else:
            return angle
        
    def intersections(self, other_arc):
        """Gives the two intersections of the greats circles defined by the 
       current arc and *other_arc*.
        """
        
        
        if self.end.lon - self.start.lon > math.pi:
            self.end.lon -= 2 * math.pi
        if other_arc.end.lon - other_arc.start.lon > math.pi:
            other_arc.end.lon -= 2 * math.pi
        if self.end.lon - self.start.lon < -math.pi:
            self.end.lon += 2 * math.pi
        if other_arc.end.lon - other_arc.start.lon < -math.pi:
            other_arc.end.lon += 2 * math.pi
            
        ea_ = self.start.cross2cart(self.end).normalize()
        eb_ = other_arc.start.cross2cart(other_arc.end).normalize()

        cross = ea_.cross(eb_)
        lat = math.atan2(cross.z__, math.sqrt(cross.x__ ** 2 + cross.y__ ** 2))
        lon = math.atan2(-cross.y__, cross.x__)

        return (Coordinate(math.degrees(lat), math.degrees(lon)),
                Coordinate(math.degrees(-lat),
                           math.degrees(modpi(lon + math.pi))))

    def intersects(self, other_arc):
        """Says if two arcs defined by the current arc and the *other_arc*
        intersect. An arc is defined as the shortest tracks between two points.
        """

        
        for i in self.intersections(other_arc):
            a__ = self.start
            b__ = self.end
            c__ = other_arc.start
            d__ = other_arc.end


            ab_ = a__.distance(b__)
            cd_ = c__.distance(d__)

            if(abs(a__.distance(i) + b__.distance(i) - ab_) < EPSILON and
               abs(c__.distance(i) + d__.distance(i) - cd_) < EPSILON):
                return True
        return False

    def intersection(self, other_arc):
        """Says where, if two arcs defined by the current arc and the
        *other_arc* intersect. An arc is defined as the shortest tracks between
        two points.
        """

        
        for i in self.intersections(other_arc):
            a__ = self.start
            b__ = self.end
            c__ = other_arc.start
            d__ = other_arc.end


            ab_ = a__.distance(b__)
            cd_ = c__.distance(d__)

            if(abs(a__.distance(i) + b__.distance(i) - ab_) < EPSILON and
               abs(c__.distance(i) + d__.distance(i) - cd_) < EPSILON):
                return i
        return None

def modpi(val):
    """Puts *val* between -pi and pi.
    """
    return (val + math.pi) % (2 * math.pi) - math.pi

def modpi2(val):
    """Puts *val* between 0 and 2pi.
    """
    return val % (2 * math.pi)


def point_inside(point, corners):
    """Is a point inside the 4 corners ? This uses great circle arcs as area
    boundaries.
    """
    arc1 = Arc(corners[0], corners[1])
    arc2 = Arc(corners[1], corners[2])
    arc3 = Arc(corners[2], corners[3])
    arc4 = Arc(corners[3], corners[0])
    
    arc5 = Arc(corners[1], point)
    arc6 = Arc(corners[3], point)

    angle1 = modpi(arc1.angle(arc2))
    angle1bis = modpi(arc1.angle(arc5))

    angle2 = modpi(arc3.angle(arc4))
    angle2bis = modpi(arc3.angle(arc6))

    return (np.sign(angle1) == np.sign(angle1bis) and
            abs(angle1) > abs(angle1bis) and 
            np.sign(angle2) == np.sign(angle2bis) and
            abs(angle2) > abs(angle2bis))
    
def overlaps(area_corners, segment_corners):
    """Are two areas overlapping ? This uses great circle arcs as area
    boundaries.
    """
    for i in area_corners:
        if point_inside(i, segment_corners):
            return True
    for i in segment_corners:
        if point_inside(i, area_corners):
            return True
    
    area_arc1 = Arc(area_corners[0], area_corners[1])
    area_arc2 = Arc(area_corners[1], area_corners[2])
    area_arc3 = Arc(area_corners[2], area_corners[3])
    area_arc4 = Arc(area_corners[3], area_corners[0])

    segment_arc1 = Arc(segment_corners[0], segment_corners[1])
    segment_arc2 = Arc(segment_corners[1], segment_corners[2])
    segment_arc3 = Arc(segment_corners[2], segment_corners[3])
    segment_arc4 = Arc(segment_corners[3], segment_corners[0])

    for i in (area_arc1, area_arc2, area_arc3, area_arc4):
        for j in (segment_arc1, segment_arc2, segment_arc3, segment_arc4):
            if i.intersects(j):
                return True
    return False

def get_intersections(b__, boundaries):
    """Get the intersections of *b__* with *boundaries*.
    Returns both the intersection coordinates and the concerned boundaries.
    """
    
    intersections = []
    bounds = []
    for other_b in boundaries:
        inter = b__.intersection(other_b)
        if inter is not None:
            intersections.append(inter)
            bounds.append(other_b)
    return intersections, bounds
    
def get_first_intersection(b__, boundaries):
    """Get the first intersection on *b__* with *boundaries*.
    """
    intersections, bounds = get_intersections(b__, boundaries)
    del bounds
    dists = np.array([b__.start.distance(p__) for p__ in intersections])
    indices = dists.argsort()
    if len(intersections) > 0:
        return intersections[indices[0]]
    return None

def get_next_intersection(p__, b__, boundaries):
    """Get the next intersection from the intersection of arcs *p__* and *b__*
    along segment *b__* with *boundaries*.
    """
    new_b = Arc(p__, b__.end)
    intersections, bounds = get_intersections(new_b, boundaries)
    dists = np.array([b__.start.distance(p2) for p2 in intersections])
    indices = dists.argsort()
    if len(intersections) > 0 and intersections[indices[0]] != p__:
        return intersections[indices[0]], bounds[indices[0]]
    elif len(intersections) > 1:
        return intersections[indices[1]], bounds[indices[1]]
    return None, None

def polygon(area_corners, segment_corners):
    """Get the intersection polygon between two areas.
    """
    area_boundaries = [Arc(area_corners[0], area_corners[1]),
                       Arc(area_corners[1], area_corners[2]),
                       Arc(area_corners[2], area_corners[3]),
                       Arc(area_corners[3], area_corners[0])]
    segment_boundaries = [Arc(segment_corners[0], segment_corners[1]),
                          Arc(segment_corners[1], segment_corners[2]),
                          Arc(segment_corners[2], segment_corners[3]),
                          Arc(segment_corners[3], segment_corners[0])]

    angle1 = area_boundaries[0].angle(area_boundaries[1])
    angle2 = segment_boundaries[0].angle(segment_boundaries[1])
    if np.sign(angle1) != np.sign(angle2):
        segment_corners.reverse()
        segment_boundaries = [Arc(segment_corners[0], segment_corners[1]),
                              Arc(segment_corners[1], segment_corners[2]),
                              Arc(segment_corners[2], segment_corners[3]),
                              Arc(segment_corners[3], segment_corners[0])]
    poly = []

    boundaries = area_boundaries
    other_boundaries = segment_boundaries

    b__ = None

    for b__ in boundaries:
        if point_inside(b__.start, segment_corners):
            poly.append(b__.start)
            break
        else:
            inter = get_first_intersection(b__, other_boundaries)
            if inter is not None:
                poly.append(inter)
                break
    if len(poly) == 0:
        return None
    while len(poly) < 2 or poly[0] != poly[-1]:
        inter, b2_ = get_next_intersection(poly[-1], b__, other_boundaries)
        if inter is None:
            poly.append(b__.end)
            idx = (boundaries.index(b__) + 1) % len(boundaries)
            b__ = boundaries[idx]
        else:
            poly.append(inter)
            b__ = b2_
            boundaries, other_boundaries = other_boundaries, boundaries
    return poly[:-1]

R = 1

def get_area(corners):
    """Get the area of the convex area defined by *corners*.
    """
    c1_ = corners[0]
    area = 0

    for idx in range(1, len(corners) - 1):
        b1_ = Arc(c1_, corners[idx])
        b2_ = Arc(c1_, corners[idx + 1])
        b3_ = Arc(corners[idx], corners[idx + 1])
        e__ = (abs(b1_.angle(b2_)) +
               abs(b2_.angle(b3_)) + 
               abs(b3_.angle(b1_)))
        area += R ** 2 * e__ - math.pi
    return area

def overlap_rate(swath_corners, area_corners):
    """Get how much a swath overlaps an area.
    """
    area_area = get_area(area_corners)
    inter_area = get_area(polygon(area_corners, swath_corners))
    return inter_area / area_area

def min_distances(area_corners, segment_corners):
    """Min distances between each corner of *area_corners* and
    *segment_corners*.
    """
    dists = np.ones(4) * np.infty
    for i, ic_ in enumerate(area_corners):
        for jc_ in segment_corners:
            dist = ic_.distance(jc_)
            if dists[i] > dist:
                dists[i] = dist
    return dists

def should_wait(area_corners, segment_corners, previous_segment_corners):
    """Are the newest cornest still inside the area ? is the last segment
    boundary overlapping any boundary of the area ? In this case we should wait
    for the next segment to arrive.
    """
    dists = min_distances(segment_corners, previous_segment_corners)
    indices = np.argsort(dists)
    new_corners = np.array(segment_corners)[indices[2:]]
    if len(new_corners) != 2:
        raise ValueError("More than 2 corners differ from previous segment...")
    new_arc = Arc(new_corners[0], new_corners[1])
    
    for i in new_corners:
        if point_inside(i, area_corners):
            return True

    area_arc1 = Arc(area_corners[0], area_corners[1])
    area_arc2 = Arc(area_corners[1], area_corners[2])
    area_arc3 = Arc(area_corners[2], area_corners[3])
    area_arc4 = Arc(area_corners[3], area_corners[0])

    for i in (area_arc1, area_arc2, area_arc3, area_arc4):
        if i.intersects(new_arc):
            return True

    return False
    
import unittest
class TestSphereGeometry(unittest.TestCase):
    """Testing sphere geometry from this module.
    """

    def test_angle(self):
        """Testing the angle value between two arcs.
        """

        base = 0

        p0_ = Coordinate(base, base)
        p1_ = Coordinate(base + 1, base)
        p2_ = Coordinate(base, base + 1)
        p3_ = Coordinate(base - 1, base)
        p4_ = Coordinate(base, base - 1)

        arc1 = Arc(p0_, p1_)
        arc2 = Arc(p0_, p2_)
        arc3 = Arc(p0_, p3_)
        arc4 = Arc(p0_, p4_)

        self.assertAlmostEqual(arc1.angle(arc2), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc2.angle(arc3), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc3.angle(arc4), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc4.angle(arc1), math.pi / 2,
                               msg="this should be pi/2")

        self.assertAlmostEqual(arc1.angle(arc4), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc4.angle(arc3), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc3.angle(arc2), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc2.angle(arc1), -math.pi / 2,
                               msg="this should be -pi/2")

        self.assertAlmostEqual(arc1.angle(arc3), math.pi,
                               msg="this should be pi")
        self.assertAlmostEqual(arc3.angle(arc1), math.pi,
                               msg="this should be pi")
        self.assertAlmostEqual(arc2.angle(arc4), math.pi,
                               msg="this should be pi")
        self.assertAlmostEqual(arc4.angle(arc2), math.pi,
                               msg="this should be pi")


        p5_ = Coordinate(base + 1, base + 1)
        p6_ = Coordinate(base - 1, base + 1)
        p7_ = Coordinate(base - 1, base - 1)
        p8_ = Coordinate(base + 1, base - 1)

        arc5 = Arc(p0_, p5_)
        arc6 = Arc(p0_, p6_)
        arc7 = Arc(p0_, p7_)
        arc8 = Arc(p0_, p8_)

        self.assertAlmostEqual(arc1.angle(arc5), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc5.angle(arc2), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc2.angle(arc6), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc6.angle(arc3), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc3.angle(arc7), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc7.angle(arc4), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc4.angle(arc8), math.pi / 4, 3,
                               msg="this should be pi/4")
        self.assertAlmostEqual(arc8.angle(arc1), math.pi / 4, 3,
                               msg="this should be pi/4")

        self.assertAlmostEqual(arc1.angle(arc6), 3 * math.pi / 4, 3,
                               msg="this should be 3pi/4")


        c0_ = Coordinate(0, 180)
        c1_ = Coordinate(1, 180)
        c2_ = Coordinate(0, -179)
        c3_ = Coordinate(-1, -180)
        c4_ = Coordinate(0, 179)


        arc1 = Arc(c0_, c1_)
        arc2 = Arc(c0_, c2_)
        arc3 = Arc(c0_, c3_)
        arc4 = Arc(c0_, c4_)

        self.assertAlmostEqual(arc1.angle(arc2), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc2.angle(arc3), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc3.angle(arc4), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc4.angle(arc1), math.pi / 2,
                               msg="this should be pi/2")

        self.assertAlmostEqual(arc1.angle(arc4), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc4.angle(arc3), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc3.angle(arc2), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc2.angle(arc1), -math.pi / 2,
                               msg="this should be -pi/2")

        # case of the north pole

        c0_ = Coordinate(90, 0)
        c1_ = Coordinate(89, 0)
        c2_ = Coordinate(89, -90)
        c3_ = Coordinate(89, 180)
        c4_ = Coordinate(89, 90)

        arc1 = Arc(c0_, c1_)
        arc2 = Arc(c0_, c2_)
        arc3 = Arc(c0_, c3_)
        arc4 = Arc(c0_, c4_)

        self.assertAlmostEqual(arc1.angle(arc2), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc2.angle(arc3), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc3.angle(arc4), math.pi / 2,
                               msg="this should be pi/2")
        self.assertAlmostEqual(arc4.angle(arc1), math.pi / 2,
                               msg="this should be pi/2")

        self.assertAlmostEqual(arc1.angle(arc4), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc4.angle(arc3), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc3.angle(arc2), -math.pi / 2,
                               msg="this should be -pi/2")
        self.assertAlmostEqual(arc2.angle(arc1), -math.pi / 2,
                               msg="this should be -pi/2")

        self.assertAlmostEqual(Arc(c1_, c2_).angle(arc1), math.pi/4, 3,
                               msg="this should be pi/4")
                               
        self.assertAlmostEqual(Arc(c4_, c3_).angle(arc4), -math.pi/4, 3,
                               msg="this should be -pi/4")

        self.assertAlmostEqual(Arc(c1_, c4_).angle(arc1), -math.pi/4, 3,
                               msg="this should be -pi/4")


    def test_inside(self):
        """Testing if a point is inside for other points.
        """
        c1_ = Coordinate(-11, -11)
        c2_ = Coordinate(11, -11)
        c3_ = Coordinate(11, 11)
        c4_ = Coordinate(-11, 11)    

        corners = [c1_, c2_, c3_, c4_]

        point = Coordinate(0, 0)
        self.assertTrue(point_inside(point, corners))

        point = Coordinate(0, 12)
        self.assertFalse(point_inside(point, corners))



        c1_ = Coordinate(-1, 180)
        c2_ = Coordinate(1, 179)
        c3_ = Coordinate(1, -179)
        c4_ = Coordinate(-1, -179)

        corners = [c1_, c2_, c3_, c4_]
        point = Coordinate(0, 180)
        self.assertTrue(point_inside(point, corners))

        point = Coordinate(12, 180)
        self.assertFalse(point_inside(point, corners))

        point = Coordinate(-12, 180)
        self.assertFalse(point_inside(point, corners))

        point = Coordinate(0, 192)
        self.assertFalse(point_inside(point, corners))

        point = Coordinate(0, -192)
        self.assertFalse(point_inside(point, corners))


        # case of the north pole
        c1_ = Coordinate(89, 0)
        c2_ = Coordinate(89, 90)
        c3_ = Coordinate(89, 180)
        c4_ = Coordinate(89, -90)    

        corners = [c1_, c2_, c3_, c4_]

        point = Coordinate(90, 90)

        self.assertTrue(point_inside(point, corners))

    def test_intersects(self):
        """Test if two arcs intersect.
        """
        p0_ = Coordinate(0, 0)
        p1_ = Coordinate(1, 0)
        p2_ = Coordinate(0, 1)
        p3_ = Coordinate(-1, 0)
        p4_ = Coordinate(0, -1)
        p5_ = Coordinate(1, 1)
        p6_ = Coordinate(-1, 1)

        arc13 = Arc(p1_, p3_)
        arc24 = Arc(p2_, p4_)

        arc32 = Arc(p3_, p2_)
        arc41 = Arc(p4_, p1_)

        arc40 = Arc(p4_, p0_)
        arc56 = Arc(p5_, p6_)

        arc45 = Arc(p4_, p5_)
        arc02 = Arc(p0_, p2_)

        arc35 = Arc(p3_, p5_)

        self.assertTrue(arc13.intersects(arc24))

        self.assertFalse(arc32.intersects(arc41))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc45.intersects(arc02))

        self.assertTrue(arc35.intersects(arc24))



        p0_ = Coordinate(0, 180)
        p1_ = Coordinate(1, 180)
        p2_ = Coordinate(0, -179)
        p3_ = Coordinate(-1, -180)
        p4_ = Coordinate(0, 179)
        p5_ = Coordinate(1, -179)
        p6_ = Coordinate(-1, -179)

        arc13 = Arc(p1_, p3_)
        arc24 = Arc(p2_, p4_)

        arc32 = Arc(p3_, p2_)
        arc41 = Arc(p4_, p1_)

        arc40 = Arc(p4_, p0_)
        arc56 = Arc(p5_, p6_)

        arc45 = Arc(p4_, p5_)
        arc02 = Arc(p0_, p2_)

        arc35 = Arc(p3_, p5_)

        self.assertTrue(arc13.intersects(arc24))

        self.assertFalse(arc32.intersects(arc41))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc45.intersects(arc02))

        self.assertTrue(arc35.intersects(arc24))

        # case of the north pole

        p0_ = Coordinate(90, 0)
        p1_ = Coordinate(89, 0)
        p2_ = Coordinate(89, 90)
        p3_ = Coordinate(89, 180)
        p4_ = Coordinate(89, -90)    
        p5_ = Coordinate(89, 45)
        p6_ = Coordinate(89, 135)

        arc13 = Arc(p1_, p3_)
        arc24 = Arc(p2_, p4_)

        arc32 = Arc(p3_, p2_)
        arc41 = Arc(p4_, p1_)

        arc40 = Arc(p4_, p0_)
        arc56 = Arc(p5_, p6_)

        arc45 = Arc(p4_, p5_)
        arc02 = Arc(p0_, p2_)

        arc35 = Arc(p3_, p5_)

        self.assertTrue(arc13.intersects(arc24))

        self.assertFalse(arc32.intersects(arc41))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc56.intersects(arc40))

        self.assertFalse(arc45.intersects(arc02))

        self.assertTrue(arc35.intersects(arc24))

    def test_overlaps(self):
        """Test if two areas overlap.
        """
        p1_ = Coordinate(89, 0)
        p2_ = Coordinate(89, 90)
        p3_ = Coordinate(89, 180)
        p4_ = Coordinate(89, -90)    
        p5_ = Coordinate(89, 45)
        p6_ = Coordinate(89, 135)
        p7_ = Coordinate(89, -135)
        p8_ = Coordinate(89, -45)

        self.assertTrue(overlaps([p1_, p2_, p3_, p4_],
                                 [p5_, p6_, p7_, p8_]))
        self.assertFalse(overlaps([p1_, p5_, p2_, p6_],
                                  [p3_, p7_, p4_, p8_]))

        p1_ = Coordinate(1, -1)
        p2_ = Coordinate(1, 1)
        p3_ = Coordinate(-1, 1)
        p4_ = Coordinate(-1, -1)

        p5_ = Coordinate(0, 0)
        p6_ = Coordinate(0, 2)
        p7_ = Coordinate(2, 2)
        p8_ = Coordinate(2, 0)

        self.assertTrue(overlaps([p1_, p2_, p3_, p4_], [p5_, p6_, p7_, p8_]))
        self.assertFalse(overlaps([p1_, p8_, p5_, p4_], [p2_, p3_, p6_, p7_]))


    def test_overlap_rate(self):
        """Test how much two areas overlap.
        """
        p1_ = Coordinate(1, -1)
        p2_ = Coordinate(1, 1)
        p3_ = Coordinate(-1, 1)
        p4_ = Coordinate(-1, -1)

        p5_ = Coordinate(0, 0)
        p6_ = Coordinate(0, 2)
        p7_ = Coordinate(2, 2)
        p8_ = Coordinate(2, 0)
        self.assertAlmostEqual(overlap_rate([p1_, p2_, p3_, p4_],
                                            [p5_, p6_, p7_, p8_]), 0.25, 3)
        
        c1_ = [(60.5944, 82.829699999999974),
               (52.859999999999999, 36.888300000000001),
               (66.7547, 2.8773),
               (80.395899999999997, 98.145499999999984)]
        c2_ = [(62.953206630716465, 7.8098183315148422),
               (62.953206630716465, 26.189349044600252),
               (53.301561187195546, 26.189349044600252),
               (53.301561187195546, 7.8098183315148422)]
        cor1 = [Coordinate(t[0], t[1]) for t in c1_]
        cor2 = [Coordinate(t[0], t[1]) for t in c2_]
        self.assertAlmostEqual(overlap_rate(cor1, cor2), 0.07, 2)

        
        c1_ = [(60.5944, 82.829699999999974),
               (52.859999999999999, 36.888300000000001),
               (66.7547, 2.8773),
               (80.395899999999997, 98.145499999999984)]
        c2_ = [(65.98228561983025, 12.108984194981202),
               (65.98228561983025, 30.490647126520301),
               (57.304862819933433, 30.490647126520301),
               (57.304862819933433, 12.108984194981202)]
        cor1 = [Coordinate(t[0], t[1]) for t in c1_]
        cor2 = [Coordinate(t[0], t[1]) for t in c2_]
        self.assertAlmostEqual(overlap_rate(cor1, cor2), 0.5, 2)

        
if __name__ == '__main__':
    unittest.main()






#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Martin.Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""PPS netcdf cloud product reader
"""

import os.path
from ConfigParser import ConfigParser
from ConfigParser import NoOptionError

from datetime import datetime
import glob
import numpy as np

import mpop.channel
from mpop import CONFIG_PATH
from mpop.plugin_base import Reader

import logging
LOG = logging.getLogger(__name__)

NEW_PRODNAMES = {'cloudtype': 'CT',
                 'cloudmask': 'CMA',
                 'precipclouds': 'PC',
                 'cpp': 'CPP',
                 'ctth': 'CTTH'}

PPS_DATASETS = ['Cloud Type',
                'Multilayer Cloud Detection',
                "SAFNWC PPS PC likelihood of intense precipitation",
                "SAFNWC PPS PC likelihood of moderate precipitation",
                "SAFNWC PPS PC likelihood of light precipitation",
                ]


class InfoObject(object):

    """Simple data and info container.
    """

    def __init__(self):
        self.info = {}
        self.data = None


class NwcSafPpsChannel(mpop.channel.GenericChannel):

    def __init__(self, filename=None):
        mpop.channel.GenericChannel.__init__(self)
        self.mda = {}
        self._projectables = []
        self._keys = []
        self._refs = {}
        self.shape = None
        if filename:
            self.read(filename)

    def read(self, filename, load_lonlat=True):
        """Read the PPS v2014 formatet data"""
        LOG.debug("New netCDF CF file format!")

        import h5py

        h5f = h5py.File(filename, 'r')
        self.mda.update(h5f.attrs.items())

        self.mda["satellite"] = h5f.attrs['platform']
        self.mda["orbit"] = h5f.attrs['orbit_number']
        try:
            self.mda["time_slot"] = datetime.strptime(h5f.attrs['time_coverage_start'][:-2],
                                                      "%Y%m%dT%H%M%S")
        except AttributeError:
            LOG.debug("No time information in product file!")

        variables = {}

        for key, item in h5f.items():
            if item.attrs.get("CLASS") != 'DIMENSION_SCALE':
                variables[key] = item

        # processed variables
        processed = set()

        non_processed = set(variables.keys()) - processed

        for var_name in non_processed:
            if var_name in ['lon', 'lat']:
                continue

            var = variables[var_name]
            if ("standard_name" not in var.attrs.keys() and
                    "long_name" not in var.attrs.keys() and
                    var.attrs.get("CLASS") != "PALETTE"):
                LOG.info("Delayed processing of " + var_name)
                continue

            # Don't know how to unambiguously decide if the array is really a
            # data array or a palette or something else!
            # FIXME!
            if "standard_name" in var.attrs.keys():
                self._projectables.append(var_name)
            elif "long_name" in var.attrs.keys():
                dset_found = False
                for item in PPS_DATASETS:
                    if var.attrs['long_name'].find(item) >= 0:
                        self._projectables.append(var_name)
                        dset_found = True
                        break
                if not dset_found:
                    self.mda[var_name] = var[:]
                    # try:
                    #     self.mda[var_name] = var[:].filled(0)
                    # except AttributeError:
                    #     self.mda[var_name] = var[:]
                    continue
            elif var.attrs.get("CLASS") == "PALETTE":
                self.mda[var_name] = var[:]
                continue

            setattr(self, var_name, InfoObject())
            for key, item in var.attrs.items():
                if key != "DIMENSION_LIST":
                    getattr(self, var_name).info[key] = item

            data = var[:]
            if 'valid_range' in var.attrs.keys():
                data = np.ma.masked_outside(data, *var.attrs['valid_range'])
            elif '_FillValue' in var.attrs.keys():
                data = np.ma.masked_where(data, var.attrs['_FillValue'])
            dataset = (data * var.attrs.get("scale_factor", 1)
                       + var.attrs.get("add_offset", 0))

            getattr(self, var_name).data = dataset

            LOG.debug("long_name: " + str(var.attrs['long_name']))
            LOG.debug("Var=" + str(var_name) + " shape=" + str(dataset.shape))

            if self.shape is None:
                self.shape = dataset.shape
            elif self.shape != dataset.shape:
                LOG.debug("Shape=" + str(dataset.shape) +
                          " Not the same shape as previous field...")
                # raise ValueError("Different variable shapes !")

            # dims = var.dimensions
            # dim = dims[0]

            processed |= set([var_name])

        non_processed = set(variables.keys()) - processed
        if len(non_processed) > 0:
            LOG.warning(
                "Remaining non-processed variables: " + str(non_processed))

        # Get lon,lat:
        # from pyresample import geometry
        # area = geometry.SwathDefinition(lons=lon, lats=lat)

        # close the h5 file.
        h5f.close()
        del h5f

        return

    def project(self, coverage):
        """Project the data"""
        LOG.debug("Projecting channel %s..." % (self.name))
        import copy
        res = copy.copy(self)

        # Project the data
        for var in self._projectables:
            LOG.info("Projecting " + str(var))
            res.__dict__[var] = copy.copy(self.__dict__[var])
            res.__dict__[var].data = coverage.project_array(
                self.__dict__[var].data)

        res.name = self.name
        res.resolution = self.resolution
        res.filled = True
        res.area = coverage.out_area
        return res

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return True
        # return len(self._projectables) > 0

    def save(self, filename, old=True, **kwargs):
        del kwargs
        if old:
            from nwcsaf_formats.ppsv2014_to_oldformat import write_product
            write_product(self, filename)

        else:
            raise NotImplementedError("Can't save to new pps format yet.")


class PPSReader(Reader):

    pformat = "h5_pps_l2"

    def load(self, satscene, *args, **kwargs):
        """Read data from file and load it into *satscene*.
        """
        lonlat_is_loaded = False

        geofilename = kwargs.get('geofilename')
        prodfilename = kwargs.get('filename')

        products = []
        if "CTTH" in satscene.channels_to_load:
            products.append("ctth")
        if "CT" in satscene.channels_to_load:
            products.append("cloudtype")
        if "CMA" in satscene.channels_to_load:
            products.append("cloudmask")
        if "PC" in satscene.channels_to_load:
            products.append("precipclouds")
        if "CPP" in satscene.channels_to_load:
            products.append("cpp")

        if len(products) == 0:
            return

        try:
            area_name = satscene.area_id or satscene.area.area_id
        except AttributeError:
            area_name = "satproj_?????_?????"

        # Looking for geolocation file

        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))

        try:
            geodir = conf.get(satscene.instrument_name + "-level3",
                              "cloud_product_geodir",
                              vars=os.environ)
        except NoOptionError:
            LOG.warning("No option 'geodir' in level3 section")
            geodir = None

        if not geofilename and geodir:
            # Load geo file from config file:
            try:
                if not satscene.orbit:
                    orbit = ""
                else:
                    orbit = satscene.orbit
                geoname_tmpl = conf.get(satscene.instrument_name + "-level3",
                                        "cloud_product_geofilename",
                                        raw=True,
                                        vars=os.environ)
                filename_tmpl = (satscene.time_slot.strftime(geoname_tmpl)
                                 % {"orbit": str(orbit).zfill(5) or "*",
                                    "area": area_name,
                                    "satellite": satscene.satname + satscene.number})

                file_list = glob.glob(os.path.join(geodir, filename_tmpl))
                if len(file_list) > 1:
                    LOG.warning("More than 1 file matching for geoloaction: "
                                + str(file_list))
                elif len(file_list) == 0:
                    LOG.warning(
                        "No geolocation file matching!: "
                        + os.path.join(geodir, filename_tmpl))
                else:
                    geofilename = file_list[0]
            except NoOptionError:
                geofilename = None

        # Reading the products

        classes = {"ctth": CloudTopTemperatureHeight,
                   "cloudtype": CloudType,
                   "cloudmask": CloudMask,
                   "precipclouds": PrecipitationClouds,
                   "cpp": CloudPhysicalProperties
                   }

        nodata_mask = False

        area = None
        lons = None
        lats = None
        chn = None
        shape = None
        read_external_geo = {}
        for product in products:
            LOG.debug("Loading " + product)

            if isinstance(prodfilename, (list, tuple, set)):
                for fname in prodfilename:
                    kwargs['filename'] = fname
                    self.load(satscene, *args, **kwargs)
                return
            elif (prodfilename and
                  os.path.basename(prodfilename).startswith('S_NWC')):
                if os.path.basename(prodfilename).split("_")[2] == NEW_PRODNAMES[product]:
                    filename = prodfilename
                else:
                    continue
            else:
                filename = conf.get(satscene.instrument_name + "-level3",
                                    "cloud_product_filename",
                                    raw=True,
                                    vars=os.environ)
                directory = conf.get(satscene.instrument_name + "-level3",
                                     "cloud_product_dir",
                                     vars=os.environ)
                pathname_tmpl = os.path.join(directory, filename)
                LOG.debug("Path = " + str(pathname_tmpl))

                if not satscene.orbit:
                    orbit = ""
                else:
                    orbit = satscene.orbit

                filename_tmpl = (satscene.time_slot.strftime(pathname_tmpl)
                                 % {"orbit": str(orbit).zfill(5) or "*",
                                    "area": area_name,
                                    "satellite": satscene.satname + satscene.number,
                                    "product": product})

                file_list = glob.glob(filename_tmpl)
                if len(file_list) == 0:
                    product_name = NEW_PRODNAMES.get(product, product)
                    LOG.info("No " + str(product) +
                             " product in old format matching")
                    filename_tmpl = (satscene.time_slot.strftime(pathname_tmpl)
                                     % {"orbit": str(orbit).zfill(5) or "*",
                                        "area": area_name,
                                        "satellite": satscene.satname + satscene.number,
                                        "product": product_name})

                    file_list = glob.glob(filename_tmpl)

                if len(file_list) > 1:
                    LOG.warning("More than 1 file matching for " + product + "! "
                                + str(file_list))
                    continue
                elif len(file_list) == 0:
                    LOG.warning(
                        "No " + product + " matching!: " + filename_tmpl)
                    continue
                else:
                    filename = file_list[0]

            chn = classes[product]()
            chn.read(filename, lonlat_is_loaded == False)
            satscene.channels.append(chn)
            # Check if geolocation is loaded:
            if not chn.area:
                read_external_geo[product] = chn
                shape = chn.shape

        # Check if some 'channel'/product needs geolocation. If some product does
        # not have geolocation, get it from the geofilename:
        if not read_external_geo:
            LOG.info("Loading PPS parameters done.")
            return

        # Load geolocation
        interpolate = False
        if geofilename:
            geodict = get_lonlat(geofilename)
            lons, lats = geodict['lon'], geodict['lat']
            if lons.shape != shape or lats.shape != shape:
                interpolate = True
                row_indices = geodict['row_indices']
                column_indices = geodict['col_indices']

            lonlat_is_loaded = True
        else:
            LOG.warning("No Geo file specified: " +
                        "Geolocation will be loaded from product")

        if lonlat_is_loaded:
            if interpolate:
                from geotiepoints import SatelliteInterpolator

                cols_full = np.arange(shape[1])
                rows_full = np.arange(shape[0])

                satint = SatelliteInterpolator((lons, lats),
                                               (row_indices,
                                                column_indices),
                                               (rows_full, cols_full))
                # satint.fill_borders("y", "x")
                lons, lats = satint.interpolate()

            try:
                from pyresample import geometry
                lons = np.ma.masked_array(lons, nodata_mask)
                lats = np.ma.masked_array(lats, nodata_mask)
                area = geometry.SwathDefinition(lons=lons,
                                                lats=lats)
            except ImportError:
                area = None

        for chn in read_external_geo.values():
            if area:
                chn.area = area
            else:
                chn.lat = lats
                chn.lon = lons

        LOG.info("Loading PPS parameters done.")

        return


class CloudType(NwcSafPpsChannel):

    def __init__(self, filename=None):
        NwcSafPpsChannel.__init__(self, filename)
        self.name = "CT"


class CloudTopTemperatureHeight(NwcSafPpsChannel):

    def __init__(self, filename=None):
        NwcSafPpsChannel.__init__(self, filename)
        self.name = "CTTH"


class CloudMask(NwcSafPpsChannel):

    def __init__(self, filename=None):
        NwcSafPpsChannel.__init__(self, filename)
        self.name = "CMA"


class PrecipitationClouds(NwcSafPpsChannel):

    def __init__(self, filename=None):
        NwcSafPpsChannel.__init__(self, filename)
        self.name = "PC"


class CloudPhysicalProperties(NwcSafPpsChannel):

    def __init__(self, filename=None):
        NwcSafPpsChannel.__init__(self, filename)
        self.name = "CPP"


def get_lonlat(filename):
    """Read lon,lat from netCDF4 CF file"""
    import h5py

    col_indices = None
    row_indices = None

    LOG.debug("Geo File = " + filename)

    h5f = h5py.File(filename, 'r')

    lon = h5f["where"]["lon"]["data"]
    no_data = h5f["where"]["lon"]["what"].attrs["nodata"]
    missing_data = h5f["where"]["lon"]["what"].attrs["missingdata"]

    lons = np.ma.masked_equal(lon[:], no_data)
    lons = np.ma.masked_equal(lons, missing_data)

    scale_factor = h5f["where"]["lon"]["what"].attrs["gain"]
    add_offset = h5f["where"]["lon"]["what"].attrs["offset"]

    lons = lons * scale_factor + add_offset

    lat = h5f["where"]["lat"]["data"]
    no_data = h5f["where"]["lat"]["what"].attrs["nodata"]
    missing_data = h5f["where"]["lat"]["what"].attrs["missingdata"]

    lats = np.ma.masked_equal(lat[:], no_data)
    lats = np.ma.masked_equal(lats, missing_data)

    scale_factor = h5f["where"]["lat"]["what"].attrs["gain"]
    add_offset = h5f["where"]["lat"]["what"].attrs["offset"]

    lats = lats * scale_factor + add_offset

    # FIXME: this is to mask out the npp bowtie deleted pixels...
    if h5f["how"].attrs['platform'] == "npp":

        new_mask = np.zeros((16, 3200), dtype=bool)
        new_mask[0, :1008] = True
        new_mask[1, :640] = True
        new_mask[14, :640] = True
        new_mask[15, :1008] = True
        new_mask[14, 2560:] = True
        new_mask[1, 2560:] = True
        new_mask[0, 2192:] = True
        new_mask[15, 2192:] = True
        new_mask = np.tile(new_mask, (lons.shape[0] / 16, 1))
        lons = np.ma.masked_where(new_mask, lons)
        lats = np.ma.masked_where(new_mask, lats)

    # close the h5 file.
    h5f.close()
    del h5f

    return {'lon': lons,
            'lat': lats,
            'col_indices': col_indices, 'row_indices': row_indices}


if __name__ == '__main__':

    from mpop.utils import debug_on
    debug_on()

    # cpp = CloudPhysicalProperties(
    #    "/data/proj/safutv/data/polar_out/direct_readout/S_NWC_CPP_eos2_66743_20141120T1143140Z_20141120T1156542Z.nc")
    ct = CloudType(
        "S_NWC_CT_noaa19_30165_20141216T0135195Z_20141216T0151114Z.h5")

    res = get_lonlat(
        "S_NWC_avhrr_noaa19_30165_20141216T0135195Z_20141216T0151114Z.h5")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the pp.geo_image module.
"""
import datetime
import unittest
import sys

import numpy as np
from mock import patch, MagicMock

# Mock some modules, so we don't need them for tests.

sys.modules['osgeo'] = MagicMock()
sys.modules['pyresample'] = MagicMock()

import mpop.imageo.geo_image as geo_image
import mpop.imageo.formats.writer_options as writer_opts


class TestGeoImage(unittest.TestCase):
    """Class for testing pp.geo_image.
    """
    def setUp(self):
        """Setup the test.
        """
        
        self.time_slot = datetime.datetime(2009, 10, 8, 14, 30)
        self.data = np.zeros((512, 512), dtype=np.uint8)
        self.img = geo_image.GeoImage(self.data,
                                      area="euro",
                                      time_slot=self.time_slot)


    @patch.object(geo_image.GeoImage, 'geotiff_save')
    def test_save(self, mock_save):
        """Save a geo image.
        """
        
        self.img.save("test.tif", compression=0)
        mock_save.assert_called_once_with("test.tif", 0, {}, None, 256,
                                          writer_options={'blocksize': 256,
                                                          'compression': 0})
        mock_save.reset_mock()
        self.img.save("test.tif", compression=9)
        mock_save.assert_called_once_with("test.tif", 9, {}, None, 256,
                                          writer_options={'blocksize': 256,
                                                          'compression': 9})
        mock_save.reset_mock()
        self.img.save("test.tif", compression=9, floating_point=True)
        mock_save.assert_called_once_with("test.tif", 9, {}, None, 256,
                                          floating_point=True,
                                          writer_options={'blocksize': 256,
                                                          'compression': 9})

        mock_save.reset_mock()
        self.img.save("test.tif", compression=9, tags={"NBITS": 20})
        mock_save.assert_called_once_with("test.tif", 9, {"NBITS": 20},
                                          None, 256,
                                          writer_options={'blocksize': 256,
                                                          'nbits': 20,
                                                          'compression': 9})
        mock_save.reset_mock()
        self.img.save("test.tif", writer_options={"compression":9})
        mock_save.assert_called_once_with("test.tif", 9, {}, None, 256,
                                          writer_options={'blocksize': 256,
                                                          'compression': 9})

        mock_save.reset_mock()
        self.img.save("test.tif", writer_options={"compression":9, "nbits":16})
        mock_save.assert_called_once_with("test.tif", 9, {"NBITS": 16},
                                          None, 256,
                                          writer_options={'blocksize': 256,
                                                          'nbits': 16,
                                                          'compression': 9})

        mock_save.reset_mock()
        self.img.save("test.tif", writer_options={"fill_value_subst": 1})
        mock_save.assert_called_once_with("test.tif", 6, {}, None, 256,
                                          writer_options={'blocksize': 256,
                                                          'compression': 6,
                                                          'fill_value_subst': 1})


        with patch.object(geo_image.Image, 'save') as mock_isave:
            self.img.save("test.png")
            mock_isave.assert_called_once_with(self.img, 'test.png', 6,
                                               fformat='png')
            mock_isave.side_effect = geo_image.UnknownImageFormat("Boom!")
            self.assertRaises(geo_image.UnknownImageFormat,
                              self.img.save, "test.dummy")


    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format.
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"

        # test with 0 compression

        raster = gtbn.return_value
        
        self.img.geotiff_save("test.tif", 0, None, {"BLA": "09"}, 256)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              2,
                                              gby,
                                              ["BLA=09",
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256',
                                               'ALPHA=YES'])
        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 255)
        self.assertTrue(mock_write_channels.call_args[0][3] is None)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1]
                               == self.data))
        
        
        dst_ds.SetGeoTransform.assert_called_once_with([1, 10, 0, 4, 0, -11])
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S")}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')

    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_compress(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format with compression.
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"

        raster = gtbn.return_value

        self.img.geotiff_save("test.tif", 9, None, None, 256)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              2,
                                              gby,
                                              ['COMPRESS=DEFLATE',
                                               'ZLEVEL=9',
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256',
                                               'ALPHA=YES'])
        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 255)
        self.assertTrue(mock_write_channels.call_args[0][3] is None)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1] == self.data))
        
        dst_ds.SetGeoTransform.assert_called_once_with([1, 10, 0, 4, 0, -11])
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S")}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')
        
    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_floats(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format with floats.
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"
        # test with floats

        raster = gtbn.return_value
        
        self.img.geotiff_save("test.tif", 0, None, None, 256,
                              floating_point=True)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              1,
                                              gf,
                                              ['TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256'])

        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 0)
        self.assertEquals(mock_write_channels.call_args[0][3], [0])
        self.assertTrue(np.all(mock_write_channels.call_args[0][1]
                               == self.data))
        
        
        dst_ds.SetGeoTransform.assert_called_once_with([1, 10, 0, 4, 0, -11])
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S")}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')

        self.fill_value = None
        self.img.mode = "RGB"
        self.assertRaises(ValueError, self.img.geotiff_save, "test.tif", 0,
                          None, None, 256,
                          floating_point=True)
        
    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_32(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff 32-bits format.
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"

        raster = gtbn.return_value
        

        self.img.geotiff_save("test.tif", 9, {"NBITS": 20}, None, 256)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              2,
                                              gui32,
                                              ['COMPRESS=DEFLATE',
                                               'ZLEVEL=9',
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256',
                                               'ALPHA=YES'])
        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 2**32 - 1)
        self.assertTrue(mock_write_channels.call_args[0][3] is None)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1] == self.data))
        
        dst_ds.SetGeoTransform.assert_called_once_with([1, 10, 0, 4, 0, -11])
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S"),
                    "NBITS": 20}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')


    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_16(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff 16-bits format.
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"

        raster = gtbn.return_value
        

        self.img.geotiff_save("test.tif", 9, {"NBITS": 15}, None, 256)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              2,
                                              gui16,
                                              ['COMPRESS=DEFLATE',
                                               'ZLEVEL=9',
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256',
                                               'ALPHA=YES'])
        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 2**16 - 1)
        self.assertTrue(mock_write_channels.call_args[0][3] is None)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1] == self.data))
        
        dst_ds.SetGeoTransform.assert_called_once_with([1, 10, 0, 4, 0, -11])
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S"),
                    "NBITS": 15}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')


    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_geotransform(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format with custom geotransform
        """
        gadr = gad.return_value
        gadr.area_extent = [1, 2, 3, 4]
        gadr.pixel_size_x = 10
        gadr.pixel_size_y = 11
        gadr.proj4_string = "+proj=geos +ellps=WGS84"
        gadr.proj_dict = {"proj": "geos", "ellps": "WGS84"}
        gadr.proj_id = "geos0"

        # test with 0 compression

        raster = gtbn.return_value
        
        self.img.geotiff_save("test.tif", 0, None, None, 256,
                              geotransform="best geotransform of the world",
                              spatialref=spaceref())
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              2,
                                              gby,
                                              ['TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256',
                                               'ALPHA=YES'])
        dst_ds = raster.Create.return_value

        #mock_write_channels.assert_called_once_with(dst_ds, self.data,
        #                                            255, None)

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 255)
        self.assertTrue(mock_write_channels.call_args[0][3] is None)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1]
                               == self.data))
        
        
        dst_ds.SetGeoTransform.assert_called_once_with("best geotransform of"
                                                       " the world")
        srs = spaceref.return_value.ExportToWkt.return_value
        dst_ds.SetProjection.assert_called_once_with(srs)

        time_tag = {"TIFFTAG_DATETIME":
                    self.img.time_slot.strftime("%Y:%m:%d %H:%M:%S")}
        dst_ds.SetMetadata.assert_called_once_with(time_tag, '')


    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_fill_value(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format.
        """
        
        # source image data, masked data but only zeros
        self.data = np.ma.zeros((512, 512), dtype=np.uint8)
        self.data.mask = np.zeros(self.data .shape, dtype=bool)
        self.data.mask[0,0] = True

        self.img = geo_image.GeoImage(self.data,
                                      area="euro",
                                      time_slot=self.time_slot)
        self.img.fill_value = [0]

        raster = gtbn.return_value
        
        self.img.geotiff_save("test.tif", 0, None, {"BLA": "09"}, 256)
        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              1,
                                              gby,
                                              ["BLA=09",
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256'])
        dst_ds = raster.Create.return_value

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 255)
        self.assertTrue(mock_write_channels.call_args[0][3], self.img.fill_value)
        self.assertTrue(np.all(mock_write_channels.call_args[0][1]
                               == self.data))


    @patch('osgeo.osr.SpatialReference')
    @patch('mpop.projector.get_area_def')
    @patch('osgeo.gdal.GDT_Float64')
    @patch('osgeo.gdal.GDT_Byte')
    @patch('osgeo.gdal.GDT_UInt16')
    @patch('osgeo.gdal.GDT_UInt32')
    @patch('osgeo.gdal.GetDriverByName')
    @patch.object(geo_image.GeoImage, '_gdal_write_channels')
    def test_save_geotiff_fill_value_subst(self, mock_write_channels, gtbn, gui32, gui16, gby, gf, gad, spaceref):
        """Save to geotiff format.
        """

        # source image data, masked data but only zeros
        self.data = np.ma.zeros((512, 512), dtype=np.uint8)
        self.data.mask = np.zeros(self.data .shape, dtype=bool)
        self.data.mask[0,0] = True

        self.img = geo_image.GeoImage(self.data,
                                      area="euro",
                                      time_slot=self.time_slot)
        self.img.fill_value = [0]
        
        # not masked zeros should be replaced by ones
        fill_value_substitution = 1
        
        data_with_subst = np.ma.copy(self.data)
        np.place(data_with_subst, self.data == self.img.fill_value[0], 1)

        raster = gtbn.return_value

        self.img.geotiff_save("test.tif", 0, None, {"BLA": "09"}, 256,
                              writer_options={writer_opts.WR_OPT_FILL_VALUE_SUBST: fill_value_substitution})

        gtbn.assert_called_once_with("GTiff")

        raster.Create.assert_called_once_with("test.tif",
                                              self.data.shape[0],
                                              self.data.shape[1],
                                              1,
                                              gby,
                                              ["BLA=09",
                                               'TILED=YES',
                                               'BLOCKXSIZE=256',
                                               'BLOCKYSIZE=256'])
        dst_ds = raster.Create.return_value

        self.assertEquals(mock_write_channels.call_count, 1)
        self.assertEquals(mock_write_channels.call_args[0][0], dst_ds)
        self.assertEquals(mock_write_channels.call_args[0][2], 255)
        self.assertTrue(mock_write_channels.call_args[0][3], self.img.fill_value)
        
        # all zeros  should be replaced by ones
        self.assertTrue(np.all(mock_write_channels.call_args[0][1]
                               == data_with_subst))


def suite():
    """The test suite for test_geo_image.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestGeoImage))
    
    return mysuite




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014, 2015.
#
# Author(s):
#
#   Panu Lahtinen <panu.lahtinen@fmi.fi>
#   Martin Raspaud <martin.raspaud@smhi.se>
#
# This file is part of mpop.
#
# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

'''Helper functions for area extent calculations.
'''

import numpy as np
from mpop.projector import get_area_def
# from pyresample.utils import AreaNotFound

import pyresample

import logging
from pyproj import Proj

LOGGER = logging.getLogger(__name__)


def area_def_names_to_extent(area_def_names, proj4_str,
                             default_extent=(-5567248.07, -5570248.48,
                                             5570248.48, 5567248.07)):
    '''Convert a list of *area_def_names* to maximal area extent in destination
    projection defined by *proj4_str*. *default_extent* gives the extreme
    values.  Default value is MSG3 extents at lat0=0.0. If a boundary of one of
    the area_defs is entirely invalid, the *default_extent* is taken.
    '''

    if type(area_def_names) is not list:
        area_def_names = [area_def_names]

    maximum_extent = None

    for name in area_def_names:

        try:
            adef = get_area_def(name)
            if "proj=geos" in adef.proj4_string:
                maximum_extent = update_max_extent(maximum_extent,
                                                   adef.area_extent)
                continue
            boundaries = adef.get_boundary_lonlats()
        except pyresample.utils.AreaNotFound:
            LOGGER.warning('Area definition not found ' + name)
            continue
        except AttributeError:
            boundaries = name.get_boundary_lonlats()

        if (any(boundaries[0].side1 > 1e20) or
                any(boundaries[0].side2 > 1e20) or
                any(boundaries[0].side3 > 1e20) or
                any(boundaries[0].side4 > 1e20)):
            if default_extent:
                maximum_extent = list(default_extent)
                continue
            else:
                return None

        lon_sides = (boundaries[0].side1, boundaries[0].side2,
                     boundaries[0].side3, boundaries[0].side4)
        lat_sides = (boundaries[1].side1, boundaries[1].side2,
                     boundaries[1].side3, boundaries[1].side4)

        maximum_extent = boundaries_to_extent(proj4_str, maximum_extent,
                                              default_extent,
                                              lon_sides, lat_sides)
        if not maximum_extent:
            return None

    maximum_extent = list(maximum_extent)
    maximum_extent[0] -= 10000
    maximum_extent[1] -= 10000
    maximum_extent[2] += 10000
    maximum_extent[3] += 10000

    return maximum_extent


def boundaries_to_extent(proj4_str, maximum_extent, default_extent,
                         lon_sides, lat_sides):
    '''Get area extent from given boundaries.
    '''

    # proj4-ify the projection string
    if '+' not in proj4_str:
        proj4_str = proj4_str.split(' ')
        proj4_str = '+' + ' +'.join(proj4_str)

    pro = Proj(proj4_str)

    # extents for edges
    x_dir, y_dir = pro(np.concatenate(lon_sides),
                       np.concatenate(lat_sides))

    # replace invalid values with NaN
    x_dir[np.abs(x_dir) > 1e20] = np.nan
    y_dir[np.abs(y_dir) > 1e20] = np.nan

    # return None when no default specified
    if not default_extent:
        if any(np.isnan(x_dir)) or any(np.isnan(x_dir)):
            return None

    # Get the maximum needed extent from different corners.
    extent = [np.nanmin(x_dir),
              np.nanmin(y_dir),
              np.nanmax(x_dir),
              np.nanmax(y_dir)]

    # Replace "infinity" values with default extent
    for i in range(4):
        if extent[i] is np.nan:
            extent[i] = default_extent[i]

    maximum_extent = update_max_extent(maximum_extent, extent)

    # Replace "infinity" values with default extent
    for i in range(4):
        if not np.isfinite(maximum_extent[i]):
            maximum_extent[i] = default_extent[i]

    return maximum_extent


def update_max_extent(maximum_extent, extent):
    """Update the *maximum_extent* to cover also *extent*"""
    # update maximum extent
    if maximum_extent is None:
        maximum_extent = extent
    else:
        if maximum_extent[0] > extent[0]:
            maximum_extent[0] = extent[0]
        if maximum_extent[1] > extent[1]:
            maximum_extent[1] = extent[1]
        if maximum_extent[2] < extent[2]:
            maximum_extent[2] = extent[2]
        if maximum_extent[3] < extent[3]:
            maximum_extent[3] = extent[3]

    return maximum_extent



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to HRPT level 0 format. Needs pyorbital.

Since the loading and calibration goes quite fast, all channels are calibrated
at the same time, so don't hesitate to load all the channels anyway.

Contrarily to AAPP, no smoothing, sigma or gross filtering is taking place.

TODO:
 - Faster navigation (pyorbital).

"""
from ConfigParser import ConfigParser
import os
import logging
import glob
import numpy as np
import numexpr as ne
from mpop.plugin_base import Reader
from mpop import CONFIG_PATH
from pyresample.geometry import SwathDefinition

logger = logging.getLogger(__name__)

# Constants
c1 = 1.1910427e-5 #mW/(m2-sr-cm-4)
c2 = 1.4387752 #cm-K 


calib = {"noaa 15": # copy from noaa 16
         {
             # VIS
             "intersections": np.array([497.5, 500.3, 498.7]),
             "slopes_l": np.array([0.0523, 0.0513, 0.0262]),
             "slopes_h": np.array([0.1528, 0.1510, 0.1920]),
             "intercepts_l": np.array([-2.016, -1.943, -1.01]),
             "intercepts_h": np.array([-51.91, -51.77, -84.2]),
             # IR
             "d0": np.array([276.355, 276.142, 275.996, 276.132, 0]),
             "d1": np.array([0.0562, 0.05605, 0.05486, 0.0594, 0]),
             "d2": np.array([-1.590e-5, -1.707e-5, -1.223e-5, -1.344e-5, 0]),
             "d3": np.array([2.486e-8, 2.595e-8, 1.862e-8, 2.112e-8, 0]),
             "d4": np.array([-1.199e-11, -1.224e-11,
                             -0.853e-11, -1.001e-11, 0]),
             "prt_weights": np.array((.25, .25, .25, .25)),
             "vc": np.array((2700.1148, 917.2289, 838.1255)),
             "A": np.array((1.592459, 0.332380, 0.674623)),
             "B": np.array((0.998147, 0.998522, 0.998363)),
             "N_S": np.array([0, -2.467, -2.009]),
             "b0": np.array([0, 2.96, 2.25]),
             "b1": np.array([0, -0.05411, -0.03665]),
             "b2": np.array([0, 0.00024532, 0.00014854]),
             },
         "noaa 16":
         {
             # VIS
             "intersections": np.array([497.5, 500.3, 498.7]),
             "slopes_l": np.array([0.0523, 0.0513, 0.0262]),
             "slopes_h": np.array([0.1528, 0.1510, 0.1920]),
             "intercepts_l": np.array([-2.016, -1.943, -1.01]),
             "intercepts_h": np.array([-51.91, -51.77, -84.2]),
             # IR
             "d0": np.array([276.355, 276.142, 275.996, 276.132, 0]),
             "d1": np.array([0.0562, 0.05605, 0.05486, 0.0594, 0]),
             "d2": np.array([-1.590e-5, -1.707e-5, -1.223e-5, -1.344e-5, 0]),
             "d3": np.array([2.486e-8, 2.595e-8, 1.862e-8, 2.112e-8, 0]),
             "d4": np.array([-1.199e-11, -1.224e-11,
                             -0.853e-11, -1.001e-11, 0]),
             "prt_weights": np.array((.25, .25, .25, .25)),
             "vc": np.array((2700.1148, 917.2289, 838.1255)),
             "A": np.array((1.592459, 0.332380, 0.674623)),
             "B": np.array((0.998147, 0.998522, 0.998363)),
             "N_S": np.array([0, -2.467, -2.009]),
             "b0": np.array([0, 2.96, 2.25]),
             "b1": np.array([0, -0.05411, -0.03665]),
             "b2": np.array([0, 0.00024532, 0.00014854]),
             },
         "noaa 18": # FIXME: copy of noaa 19
         {
             # VIS
             "intersections": np.array([496.43, 500.37, 496.11]),
             "slopes_l": np.array([0.055091, 0.054892, 0.027174]),
             "slopes_h": np.array([0.16253, 0.16325, 0.18798]),
             "intercepts_l": np.array([-2.1415, -2.1288, -1.0881]),
             "intercepts_h": np.array([-55.863, -56.445, -81.491]),
             # IR
             "d0": np.array([276.601, 276.683, 276.565, 276.615, 0]),
             "d1": np.array([0.05090, 0.05101, 0.05117, 0.05103, 0]),
             "d2": np.array([1.657e-6, 1.482e-6, 1.313e-6, 1.484e-6, 0]),
             "d3": np.array([0, 0, 0, 0, 0]),
             "d4": np.array([0, 0, 0, 0, 0]),
             "prt_weights": np.array((1, 1, 1, 1)),
             "vc": np.array((2659.7952, 928.1460, 833.2532)),
             "A": np.array((1.698704, 0.436645, 0.253179)),
             "B": np.array((0.996960, 0.998607, 0.999057)),
             "N_S": np.array([0, -5.49, -3.39]),
             "b0": np.array([0, 5.70, 3.58]),
             "b1": np.array([0, -0.11187, -0.05991]),
             "b2": np.array([0, 0.00054668, 0.00024985]),
             },
         "noaa 19":
         {
             # VIS
             "intersections": np.array([496.43, 500.37, 496.11]),
             "slopes_l": np.array([0.055091, 0.054892, 0.027174]),
             "slopes_h": np.array([0.16253, 0.16325, 0.18798]),
             "intercepts_l": np.array([-2.1415, -2.1288, -1.0881]),
             "intercepts_h": np.array([-55.863, -56.445, -81.491]),
             # IR
             "d0": np.array([276.601, 276.683, 276.565, 276.615, 0]),
             "d1": np.array([0.05090, 0.05101, 0.05117, 0.05103, 0]),
             "d2": np.array([1.657e-6, 1.482e-6, 1.313e-6, 1.484e-6, 0]),
             "d3": np.array([0, 0, 0, 0, 0]),
             "d4": np.array([0, 0, 0, 0, 0]),
             "prt_weights": np.array((1, 1, 1, 1)),
             "vc": np.array((2659.7952, 928.1460, 833.2532)),
             "A": np.array((1.698704, 0.436645, 0.253179)),
             "B": np.array((0.996960, 0.998607, 0.999057)),
             "N_S": np.array([0, -5.49, -3.39]),
             "b0": np.array([0, 5.70, 3.58]),
             "b1": np.array([0, -0.11187, -0.05991]),
             "b2": np.array([0, 0.00054668, 0.00024985]),
             },
         "metop-a":
         {
             # VIS
             "intersections": np.array([501, 500, 502]),
             "slopes_l": np.array([0.0537, 0.0545, 0.0264]),
             "slopes_h": np.array([0.1587, 0.1619, 0.1837]),
             "intercepts_l": np.array([-2.1719, -2.167, -1.0868]),
             "intercepts_h": np.array([-54.7824, -55.913, -80.0116]),
             # IR
             "d0": np.array([276.6194, 276.6511, 276.6597, 276.3685, 0]),
             "d1": np.array([0.050919, 0.050892, 0.050845, 0.050992, 0]),
             "d2": np.array([1.470892e-6, 1.489e-6, 1.520646e-6, 1.48239e-6, 0]),
             "d3": np.array([0, 0, 0, 0, 0]),
             "d4": np.array([0, 0, 0, 0, 0]),
             "prt_weights": np.array((1, 1, 1, 1)) / 4.0,
             "vc": np.array((2687, 927.2, 837.7)),
             "A": np.array((2.06699, 0.55126, 0.34716)),
             "B": np.array((0.996577, 0.998509, 0.998947)),
             "N_S": np.array([0, -4.98, -3.40]),
             "b0": np.array([0, 5.44, 3.84]),
             "b1": np.array([0, 0.89848 - 1, 0.93751 - 1]),
             "b2": np.array([0, 0.00046964, 0.00025239]),
             },
         "metop-b":
         {
             # VIS
             "intersections": np.array([501, 503, 501]),
             "slopes_l": np.array([0.053572113, 0.051817433, 0.023518528]),
             "slopes_h": np.array([0.15871941, 0.15264062, 0.16376181]),
             "intercepts_l": np.array([-2.1099778, -2.0923391, -0.9879577]),
             "intercepts_h": np.array([-54.751018, -52.806460, -71.229881]),
             # IR
             "d0": np.array([276.5853, 276.5335, 276.5721, 276.5750, 0]),
             "d1": np.array([0.050933, 0.051033, 0.051097, 0.05102, 0]),
             "d2": np.array([1.54333e-6, 1.49751e-6, 1.42928e-6, 1.50841e-6, 0]),
             "d3": np.array([0, 0, 0, 0, 0]),
             "d4": np.array([0, 0, 0, 0, 0]),
             "prt_weights": np.array((1, 1, 1, 1)) / 4.0,
             "vc": np.array((2687, 927.2, 837.7)),
             "A": np.array((2.06699, 0.55126, 0.34716)),
             "B": np.array((0.996577, 0.998509, 0.998947)),
             "N_S": np.array([0, -4.75, -4.39]),
             "b0": np.array([0, 4.85, 4.36]),
             "b1": np.array([0, 0.903229 - 1, 0.923365 - 1]),
             "b2": np.array([0, 0.00048091, 0.00033524]),
             }

         }

SATELLITES = {7: "noaa 15",
              3: "noaa 16",
              5: "noaa 18",
              13: "noaa 18",
              15: "noaa 19"}

def bfield(array, bit):
    """return the bit array.
    """
    return (array & 2**(9 - bit + 1)).astype(np.bool)  

class HRPTReader(Reader):
    """HRPT minor frame reader.
    """
    pformat = "hrpt_hmf"
    
    def load(self, satscene):
        """Read data from file and load it into *satscene*.
        """
        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
        options = {}
        for option, value in conf.items(satscene.instrument_name + "-level2",
                                        raw = True):
            options[option] = value
        CASES[satscene.instrument_name](self, satscene, options)

    def load_avhrr(self, satscene, options):
        """Read avhrr data from file and load it into *satscene*.
        """

        if "filename" not in options:
            raise IOError("No filename given, cannot load.")
        filename = os.path.join(
            options["dir"],
            (satscene.time_slot.strftime(options["filename"])))

        file_list = glob.glob(satscene.time_slot.strftime(filename))

        if len(file_list) > 1:
            raise IOError("More than one hrpt file matching!")
        elif len(file_list) == 0:
            raise IOError("No hrpt file matching!: " +
                          satscene.time_slot.strftime(filename))


        filename = file_list[0]
        array = read_file(filename)

        sat = (array["id"]["id"] & (2 ** 6 - 1)) >> 3
        sat = SATELLITES[sat[len(sat) / 2]]

        lon, lat, alt = navigate(array["timecode"], sat)
        area = SwathDefinition(lon.reshape(2048, -1), lat.reshape(2048, -1))
        satscene.area = area

        vis = vis_cal(array["image_data"][:, :, :3], sat)
        ir_ = ir_cal(array["image_data"][:, :, 2:], array["telemetry"]["PRT"],
                     array["back_scan"], array["space_data"], sat)

        channels = np.empty(array["image_data"].shape, dtype=np.float64)
        channels[:, :, :2] = vis[:, :, :2]
        channels[:, :, 3:] = ir_[:, :, 1:]
        ch3a = bfield(array["id"]["id"], 10)
        ch3b = np.logical_not(ch3a)
        channels[ch3a, :, 2] = vis[ch3a, :, 2]
        channels[ch3b, :, 2] = ir_[ch3b, :, 0]

        for chan in satscene.channels_to_load:
            if chan == "1":
                satscene["1"] = np.ma.array(vis[:, :, 0])
            if chan == "2":
                satscene["2"] = np.ma.array(vis[:, :, 1])
            if chan == "3A":
                satscene["3A"] = np.ma.array(vis[:, :, 2],
                                             mask=np.tile(ch3a, (1, 2048)))
            if chan == "3B":
                satscene["3B"] = np.ma.array(ir_[:, :, 0],
                                             mask=np.tile(ch3b, (1, 2048)))
            if chan == "4":
                satscene["4"] = np.ma.array(ir_[:, :, 1])
            if chan == "5":
                satscene["5"] = np.ma.array(ir_[:, :, 2])

## Reading
## http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c4/sec4-1.htm#t413-1

def read_file(filename):
    """Read the file using numpy
    """
    dtype = np.dtype([('frame_sync', '>u2', (6, )),
                      ('id', [('id', '>u2'),
                              ('spare', '>u2')]),
                      ('timecode', '>u2', (4, )),
                      ('telemetry', [("ramp_calibration", '>u2', (5, )),
                                     ("PRT", '>u2', (3, )),
                                     ("ch3_patch_temp", '>u2'),
                                     ("spare", '>u2'),]),
                      ('back_scan', '>u2', (10, 3)),
                      ('space_data', '>u2', (10, 5)),
                      ('sync', '>u2'),
                      ('TIP_data', '>u2', (520, )),
                      ('spare', '>u2', (127, )),
                      ('image_data', '>u2', (2048, 5)),
                      ('aux_sync', '>u2', (100, ))])

    arr = np.memmap(filename, dtype=dtype)
    #arr = arr.newbyteorder()
    return arr

## navigation

from pyorbital.orbital import Orbital
from datetime import datetime, timedelta, time
from pyorbital.geoloc import ScanGeometry, compute_pixels, get_lonlatalt

def timecode(tc_array):
    word = tc_array[0]
    day = word >> 1
    word = tc_array[1]
    msecs = ((127) & word) * 1024
    word = tc_array[2]
    msecs += word & 1023
    msecs *= 1024
    word = tc_array[3]
    msecs += word & 1023
    return datetime(2014, 1, 1) + timedelta(days=int(day) - 1,
                                            milliseconds=int(msecs))
def navigate(timecodes, satellite):
    orb = Orbital(satellite)


    first_time = timecode(timecodes[0])
    first_time = datetime(first_time.year, first_time.month, first_time.day)

    hrpttimes = [timecode(x) - first_time for x in timecodes]
    hrpttimes = np.array([x.seconds + x.microseconds / 1000000.0
                          for x in hrpttimes])
    
    scan_points = np.arange(2048)
    if satellite == "noaa 16":
        scan_angle = 55.25
    else:
        scan_angle = 55.37
    scans_nb = len(hrpttimes)
    avhrr_inst = np.vstack(((scan_points / 1023.5 - 1)
                            * np.deg2rad(-scan_angle),
                            np.zeros((len(scan_points),)))).T
    avhrr_inst = np.tile(avhrr_inst, [scans_nb, 1])

    offset = hrpttimes

    times = (np.tile(scan_points * 0.000025, [scans_nb, 1])
             + np.expand_dims(offset, 1))

    sgeom = ScanGeometry(avhrr_inst, times.ravel())

    s_times = sgeom.times(first_time)

    rpy = (0, 0, 0)
    pixels_pos = compute_pixels((orb.tle._line1, orb.tle._line2), sgeom, s_times, rpy)
    pos_time = get_lonlatalt(pixels_pos, s_times)
    return pos_time
    
    

## VIS calibration

def vis_cal(vis_data, sat):
    """Calibrates the visual data using dual gain.
    """
    logger.debug("Visual calibration")
    vis = np.empty(vis_data.shape, dtype=np.float64)
    for i in range(3):
        ch = vis_data[:, :, i]
        intersect = calib[sat]["intersections"][i]
        slope_l = calib[sat]["slopes_l"][i]
        slope_h = calib[sat]["slopes_h"][i]
        intercept_l = calib[sat]["intercepts_l"][i]
        intercept_h = calib[sat]["intercepts_h"][i]

        vis[:, :, i] = ne.evaluate("where(ch < intersect, ch * slope_l + intercept_l, ch * slope_h + intercept_h)")
    return vis

## IR calibration
def ir_cal(ir_data, telemetry, back_scan, space_data, sat):
    alen = ir_data.shape[0]
    logger.debug("IR calibration")
    logger.debug(" Preparing telemetry...")
    factor = np.ceil(alen / 5.0) + 1

    displacement = (telemetry[0:5, :] == np.array([0, 0, 0])).sum(1).argmax() + 1
    offset = 4 - (displacement - 1)

    globals().update(calib[sat])
    
    bd0 = np.tile(d0.reshape(-1, 1), (factor, 3))[offset:offset + alen]
    bd1 = np.tile(d1.reshape(-1, 1), (factor, 3))[offset:offset + alen]
    bd2 = np.tile(d2.reshape(-1, 1), (factor, 3))[offset:offset + alen]
    bd3 = np.tile(d3.reshape(-1, 1), (factor, 3))[offset:offset + alen]
    bd4 = np.tile(d4.reshape(-1, 1), (factor, 3))[offset:offset + alen]

    PRT = telemetry

    T_PRT = bd0 + PRT * (bd1 + PRT * (bd2 + PRT * (bd3 + PRT * bd4)))

    sublen = np.floor((T_PRT.shape[0] - displacement) / 5.0) * 5
    TMP_PRT = T_PRT[displacement:displacement + sublen]

    logger.debug(" Computing blackbody temperatures...")
    
    MEAN = ((TMP_PRT[::5] +
             TMP_PRT[1::5] +
             TMP_PRT[2::5] +
             TMP_PRT[3::5]) / 4).repeat(5, 0)

    if displacement == 0:
        T_BB_beg = None
    elif displacement == 1:
        T_BB_beg = MEAN[0]
    else:
        T_BB_beg = np.tile(T_PRT[:displacement].sum(0) / (displacement - 1), (displacement, 1))
    if sublen + displacement >=T_PRT.shape[0]:
        T_BB_end = None
    else:
        T_BB_end = np.tile(T_PRT[sublen+displacement:].mean(0), (T_PRT.shape[0] - sublen - displacement, 1))

    if T_BB_beg is not None:
        to_stack = [T_BB_beg, MEAN]
    else:
        to_stack = [MEAN]

    if T_BB_end is not None:
        to_stack.append(T_BB_end)

    T_BB = np.vstack(to_stack)

    if sat in ["noaa 15", "noaa 16"]:
        # three readings for klm
        T_BB = T_BB.mean(0)

    T_BB_star = A + B * T_BB

    N_BB = (c1 * vc ** 3) / (np.exp((c2 * vc)/(T_BB_star)) - 1)

    C_S = space_data[:,:, 2:].mean(1)
    C_BB = back_scan.mean(1)

    C_E = ir_data

    # aapp style
    #G = (N_BB - N_S) / (C_BB - C_S)
    #k1 = G**2 * b2
    #k2 = (b1 + 1) *G - 2 * k1 * C_S + 2*b2 * G * N_S
    #k3 = b0 + (b1 + 1) * N_S - (b1 + 1) *G * C_S + b2 * (N_S - G * N_S) ** 2
    #N_E = k1[:, np.newaxis, :] * C_E * C_E + k2[:, np.newaxis, :] * C_E + k3[:, np.newaxis, :]


    logger.debug(" Computing linear part of radiances...")

    C_Sr = C_S[:, np.newaxis, :]
    Cr = ((N_BB - N_S) / (C_S - C_BB))[:, np.newaxis, :]
    N_lin = ne.evaluate("(N_S + (Cr * (C_Sr - C_E)))")

    logger.debug(" Computing radiance correction...")
    # the +1 (N_lin) here is for Ne = Nlin + Ncor 
    N_E = ne.evaluate("((b2 * N_lin + b1 + 1) * N_lin + b0)")

    logger.debug(" Computing channels brightness temperatures...")
    T_E_star = ne.evaluate("(c2 * vc / (log(1 + c1 * vc**3 / N_E)))")
    T_E = ne.evaluate("(T_E_star - A) / B")

    return T_E


CASES = {
    "avhrr": HRPTReader.load_avhrr
    }


if __name__ == '__main__':
    import sys
    array = read_file(sys.argv[1])
    sat = (array["id"]["id"] & (2 ** 6 - 1)) >> 3
    sat = int(np.round(np.mean(sat)))
    sat = SATELLITES[sat]

    vis = vis_cal(array["image_data"][:, :, :3], sat)
    ir_ = ir_cal(array["image_data"][:, :, 2:], array["telemetry"]["PRT"],
                 array["back_scan"], array["space_data"], sat)

    channels = np.empty(array["image_data"].shape, dtype=np.float64)
    channels[:, :, :2] = vis[:, :, :2]
    channels[:, :, 3:] = ir_[:, :, 1:]
    ch3a = bfield(array["id"]["id"], 10)
    ch3b = np.logical_not(ch3a)
    channels[ch3a, :, 2] = vis[ch3a, :, 2]
    channels[ch3b, :, 2] = ir_[ch3b, :, 0]

    lon, lat, alt = navigate(array["timecode"], sat)
    area = SwathDefinition(lon.reshape(2048, -1), lat.reshape(2048, -1))




"""Loader for MSG, nwcsaf high resolution hdf5 format.
"""
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import os
from numpy import array as np_array
from numpy import empty as np_empty
from numpy import append as np_append
from numpy import dtype as np_dtype
from numpy import append as np_append
from numpy import where as np_where
from numpy import in1d as np_in1d
from numpy import logical_and as np_logical_and
from glob import glob
from mpop.projector import get_area_def
import datetime 

from copy import deepcopy

try:
    import h5py
except ImportError:
    print "... module h5py needs to be installed"
    quit()

from mipp.xrit.MSG import _Calibrator

import logging
LOG = logging.getLogger(__name__)
#from mpop.utils import debug_on
#debug_on()

GP_IDs = { 321: '08',  # Meteosat 8 
           322: '09',  # Meteosat 9 
           323: '10',  # Meteosat 10
           324: '11' } # Meteosat 11

dict_channel = {'CHANNEL00':'HRV',   'CHANNEL01':'VIS006','CHANNEL02':'VIS008','CHANNEL03':'IR_016','CHANNEL04':'IR_039','CHANNEL05':'WV_062',\
                'CHANNEL06':'WV_073','CHANNEL07':'IR_087','CHANNEL08':'IR_097','CHANNEL09':'IR_108','CHANNEL10':'IR_120','CHANNEL11':'IR_134'}


# class definition of a high resolution wind data
class HRW_class:

    def __init__(self):
        # see http://docs.scipy.org/doc/numpy/reference/generated/numpy.dtype.html
        self.date            = None # datetime of the observation  
        self.detailed        = None # False-> basic, True -> detailed 
        self.channel         = np_array([], dtype='|S6') 
        self.wind_id         = np_array([], dtype=int)
        self.prev_wind_id    = np_array([], dtype=int)
        self.segment_X       = np_array([], dtype='f') 
        self.segment_Y       = np_array([], dtype='f') 
        self.t_corr_method   = np_array([], dtype=int) 
        self.lon             = np_array([], dtype='f')     #  6.3862 [longitude in degree E]
        self.lat             = np_array([], dtype='f')     # 46.8823 [latitude in degree N]
        self.dlon            = np_array([], dtype='f')     # -0.011  [longitude in degree E]
        self.dlat            = np_array([], dtype='f')     #  0.01   [latitude in degree N]
        self.pressure        = np_array([], dtype='f')     # 64200.0 [p in Pa]
        self.wind_speed      = np_array([], dtype='f')     #   3.1   [v in m/s]
        self.wind_direction  = np_array([], dtype='f')     #  313.0  [v_dir in deg]
        self.temperature     = np_array([], dtype='f')     #  272.4  [T in K]
        self.conf_nwp        = np_array([], dtype='f')
        self.conf_no_nwp     = np_array([], dtype='f')
        self.t_type          = np_array([], dtype=int)
        self.t_level_method  = np_array([], dtype=int)
        self.t_winds         = np_array([], dtype=int)
        self.t_corr_test     = np_array([], dtype=int)
        self.applied_QI      = np_array([], dtype=int)
        self.NWP_wind_levels = np_array([], dtype=int)
        self.num_prev_winds  = np_array([], dtype=int)
        self.orographic_index= np_array([], dtype=int)
        self.cloud_type      = np_array([], dtype=int)
        self.wind_channel    = np_array([], dtype=int)
        self.correlation     = np_array([], dtype=int)
        self.pressure_error  = np_array([], dtype='f')

    # ---------------- add two data sets e.g. time steps ---------------------
    def __add__(self, HRW_class2):

        HRW_new = HRW_class()

        HRW_new.date            = self.date      # !!! does not make sense !!! 
        HRW_new.detailed        = self.detailed  # !!! does not make sense !!!
        HRW_new.channel         = np_append(self.channel,         HRW_class2.channel)
        HRW_new.wind_id         = np_append(self.wind_id,         HRW_class2.wind_id)
        HRW_new.prev_wind_id    = np_append(self.prev_wind_id,    HRW_class2.prev_wind_id)
        HRW_new.segment_X       = np_append(self.segment_X,       HRW_class2.segment_X)
        HRW_new.segment_Y       = np_append(self.segment_Y,       HRW_class2.segment_Y)
        HRW_new.t_corr_method   = np_append(self.t_corr_method,   HRW_class2.t_corr_method)
        HRW_new.lon             = np_append(self.lon,             HRW_class2.lon)
        HRW_new.lat             = np_append(self.lat,             HRW_class2.lat)
        HRW_new.dlon            = np_append(self.dlon,            HRW_class2.dlon)
        HRW_new.dlat            = np_append(self.dlat,            HRW_class2.dlat)
        HRW_new.pressure        = np_append(self.pressure,        HRW_class2.pressure)
        HRW_new.wind_speed      = np_append(self.wind_speed,      HRW_class2.wind_speed)
        HRW_new.wind_direction  = np_append(self.wind_direction,  HRW_class2.wind_direction)       
        HRW_new.temperature     = np_append(self.temperature,     HRW_class2.temperature)
        HRW_new.conf_nwp        = np_append(self.conf_nwp,        HRW_class2.conf_nwp)
        HRW_new.conf_no_nwp     = np_append(self.conf_no_nwp,     HRW_class2.conf_no_nwp)
        HRW_new.t_type          = np_append(self.t_type,          HRW_class2.t_type)
        HRW_new.t_level_method  = np_append(self.t_level_method,  HRW_class2.t_level_method)
        HRW_new.t_winds         = np_append(self.t_winds,         HRW_class2.t_winds)
        HRW_new.t_corr_test     = np_append(self.t_corr_test,     HRW_class2.t_corr_test)
        HRW_new.applied_QI      = np_append(self.applied_QI,      HRW_class2.applied_QI)
        HRW_new.NWP_wind_levels = np_append(self.NWP_wind_levels, HRW_class2.NWP_wind_levels)
        HRW_new.num_prev_winds  = np_append(self.num_prev_winds,  HRW_class2.num_prev_winds)
        HRW_new.orographic_index= np_append(self.orographic_index,HRW_class2.orographic_index)
        HRW_new.cloud_type      = np_append(self.cloud_type,      HRW_class2.cloud_type)
        HRW_new.wind_channel    = np_append(self.wind_channel,    HRW_class2.wind_channel)
        HRW_new.correlation     = np_append(self.correlation,     HRW_class2.correlation)
        HRW_new.pressure_error  = np_append(self.pressure_error,  HRW_class2.pressure_error)

        return HRW_new

    # ---------------- filter for certain criterias  ---------------------
    def filter(self, **kwargs):

        # if empty than return self (already empty)
        if self.channel.size == 0:
            return self

        HRW_new = deepcopy(self)

        for key_filter in ['min_correlation', 'min_conf_nwp', 'min_conf_no_nwp', 'cloud_type', 'level']:
            if key_filter in kwargs.keys():
                
                # if argument given is None or all keyword than skip this filter 
                if kwargs[key_filter] == None or kwargs[key_filter] == 'all' or kwargs[key_filter] == 'ALL' or kwargs[key_filter] == 'A':
                    continue

                n1 = str(HRW_new.channel.size)

                if key_filter == 'min_correlation':
                    inds = np_where(HRW_new.correlation > kwargs[key_filter])
                elif key_filter == 'min_conf_nwp':
                    inds = np_where(HRW_new.conf_nwp    > kwargs[key_filter])
                elif key_filter == 'min_conf_no_nwp':
                    inds = np_where(HRW_new.conf_no_nwp > kwargs[key_filter])
                elif key_filter == 'cloud_type':
                    mask = np_in1d(HRW_new.cloud_type, kwargs[key_filter]) 
                    inds = np_where(mask)[0]
                elif key_filter == 'level':
                    if kwargs[key_filter] == 'H': # high level: < 440hPa like in the ISCCP
                        inds = np_where(HRW_new.pressure < 44000 ) 
                    elif kwargs[key_filter] == 'M': # mid level: 440hPa ... 680hPa like in the ISCCP
                        inds = np_where( np_logical_and(44000 < HRW_new.pressure, HRW_new.pressure < 68000) ) 
                    elif kwargs[key_filter] == 'L': # low level: > 680hPa like in the ISCCP
                        inds = np_where(68000 < HRW_new.pressure)

                HRW_new.subset(inds)
                print "    filter for "+key_filter+" = ", kwargs[key_filter],' ('+n1+'->'+str(HRW_new.channel.size)+')'

        return HRW_new

    # ---------------- reduce HRW_dataset to the given indices inds ---------------------
    def subset(self, inds):

        self.channel          = self.channel         [inds]    
        self.wind_id          = self.wind_id         [inds]    
        self.prev_wind_id     = self.prev_wind_id    [inds]    
        self.segment_X        = self.segment_X       [inds]   
        self.segment_Y        = self.segment_Y       [inds]   
        self.t_corr_method    = self.t_corr_method   [inds]   
        self.lon              = self.lon             [inds]   
        self.lat              = self.lat             [inds]   
        self.dlon             = self.dlon            [inds]  
        self.dlat             = self.dlat            [inds]   
        self.pressure         = self.pressure        [inds]   
        self.wind_speed       = self.wind_speed      [inds]   
        self.wind_direction   = self.wind_direction  [inds]  
        self.temperature      = self.temperature     [inds]   
        self.conf_nwp         = self.conf_nwp        [inds]   
        self.conf_no_nwp      = self.conf_no_nwp     [inds]   
        self.t_type           = self.t_type          [inds]  
        self.t_level_method   = self.t_level_method  [inds]  
        self.t_winds          = self.t_winds         [inds] 
        self.t_corr_test      = self.t_corr_test     [inds]  
        self.applied_QI       = self.applied_QI      [inds]  
        self.NWP_wind_levels  = self.NWP_wind_levels [inds] 
        self.num_prev_winds   = self.num_prev_winds  [inds]
        self.orographic_index = self.orographic_index[inds]
        self.cloud_type       = self.cloud_type      [inds]
        self.wind_channel     = self.wind_channel    [inds]
        self.correlation      = self.correlation     [inds]
        self.pressure_error   = self.pressure_error  [inds]

        return self


def load(satscene, calibrate=True, area_extent=None, read_basic_or_detailed='both', **kwargs):
    """Load MSG SEVIRI High Resolution Wind (HRW) data from hdf5 format.
    """

    # Read config file content
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    values = {"orbit": satscene.orbit,
    "satname": satscene.satname,
    "number": satscene.number,
    "instrument": satscene.instrument_name,
    "satellite": satscene.fullname
    }

    LOG.info("assume seviri-level5")
    print "... assume seviri-level5"

    satscene.add_to_history("hdf5 data read by mpop/nwcsaf_hrw_hdf.py")

    # end of scan time 4 min after start 
    end_time = satscene.time_slot + datetime.timedelta(minutes=4)

    # area !!! satscene.area

    filename = os.path.join( satscene.time_slot.strftime(conf.get("seviri-level5", "dir", raw=True)),
                             satscene.time_slot.strftime(conf.get("seviri-level5", "filename", raw=True)) % values )

    # define classes before we search for files (in order to return empty class if no file is found)
    HRW_basic             = HRW_class()
    HRW_basic.detailed    = False 
    HRW_basic.date        = satscene.time_slot
    HRW_detailed          = HRW_class()
    HRW_detailed.detailed = True
    HRW_detailed.date     = satscene.time_slot

    print "... search for file: ", filename
    filenames=glob(str(filename))

    if len(filenames) != 0:

        if len(filenames) > 1:
            print "*** Warning, more than 1 datafile found: ", filenames 

        filename = filenames[0]
        print("... read data from %s" % str(filename))

        # create an instant of the HRW_class
        m_per_s_to_knots = 1.944

        ## limit channels to read 
        #hrw_channels=['HRV']
        # limit basic or detailed or both
        #read_basic_or_detailed='detailed'
        #read_basic_or_detailed='basic'


        with h5py.File(filename,'r') as hf:

            #print hf.attrs.keys()
            #print hf.attrs.values()

            region_name = hf.attrs['REGION_NAME'].replace("_", "")
            print "... read HRW data for region ", region_name
            LOG.info("... read HRW data for region "+region_name)
            sat_ID = GP_IDs[int(hf.attrs["GP_SC_ID"])]
            print "... derived from Meteosat ", sat_ID
            LOG.info("... derived from Meteosat "+sat_ID)

            # print('List of arrays in this file: \n', hf.keys()), len(hf.keys())

            if len(hf.keys()) == 0:
                print "*** Warning, empty file ", filename
                print ""
            else:
                for key in hf.keys():

                    if key[4:9] == "BASIC":
                        if 'read_basic_or_detailed' in locals():
                            if read_basic_or_detailed.lower() == "detailed":
                                continue
                        HRW_data = HRW_basic   # shallow copy 
                    elif key[4:12] == "DETAILED":
                        if 'read_basic_or_detailed' in locals():
                            if read_basic_or_detailed.lower() == "basic":
                                continue
                        HRW_data = HRW_detailed # shallow copy 

                    hrw_chn = dict_channel[key[len(key)-9:]]

                    if 'hrw_channels' in locals():
                        if hrw_channels != None:
                            if hrw_chn not in hrw_channels:
                                print "... "+hrw_chn+" is not in hrw_channels", hrw_channels 
                                print "    skip reading this channel" 
                                continue 

                    # read all  data 
                    channel = hf.get(key)
                    # print '... read wind vectors of channel ', channel.name, hrw_chn
                    # print  "  i    lon        lat      speed[kn] dir   pressure"
                    #for i in range(channel.len()):
                    #    print '%3d %10.7f %10.7f %7.2f %7.1f %8.1f' % (channel[i]['wind_id'], channel[i]['lon'], channel[i]['lat'], \
                    #                                                   channel[i]['wind_speed']*m_per_s_to_knots, \
                    #                                                   channel[i]['wind_direction'], channel[i]['pressure'])
                    # create string array with channel names 
                    channel_chararray = np_empty(channel.len(), dtype='|S6')
                    channel_chararray[:] = hrw_chn

                    HRW_data.channel          = np_append(HRW_data.channel         , channel_chararray              )
                    HRW_data.wind_id          = np_append(HRW_data.wind_id         , channel[:]['wind_id']          )    
                    HRW_data.prev_wind_id     = np_append(HRW_data.prev_wind_id    , channel[:]['prev_wind_id']     )    
                    HRW_data.segment_X        = np_append(HRW_data.segment_X       , channel[:]['segment_X']        )   
                    HRW_data.segment_Y        = np_append(HRW_data.segment_Y       , channel[:]['segment_Y']        )   
                    HRW_data.t_corr_method    = np_append(HRW_data.t_corr_method   , channel[:]['t_corr_method']    )   
                    HRW_data.lon              = np_append(HRW_data.lon             , channel[:]['lon']              )   
                    HRW_data.lat              = np_append(HRW_data.lat             , channel[:]['lat']              )   
                    HRW_data.dlon             = np_append(HRW_data.dlon            , channel[:]['dlon']             )  
                    HRW_data.dlat             = np_append(HRW_data.dlat            , channel[:]['dlat']             )   
                    HRW_data.pressure         = np_append(HRW_data.pressure        , channel[:]['pressure']         )   
                    HRW_data.wind_speed       = np_append(HRW_data.wind_speed      , channel[:]['wind_speed']       )   
                    HRW_data.wind_direction   = np_append(HRW_data.wind_direction  , channel[:]['wind_direction']   )  
                    HRW_data.temperature      = np_append(HRW_data.temperature     , channel[:]['temperature']      )   
                    HRW_data.conf_nwp         = np_append(HRW_data.conf_nwp        , channel[:]['conf_nwp']         )   
                    HRW_data.conf_no_nwp      = np_append(HRW_data.conf_no_nwp     , channel[:]['conf_no_nwp']      )   
                    HRW_data.t_type           = np_append(HRW_data.t_type          , channel[:]['t_type']           )  
                    HRW_data.t_level_method   = np_append(HRW_data.t_level_method  , channel[:]['t_level_method']   )  
                    HRW_data.t_winds          = np_append(HRW_data.t_winds         , channel[:]['t_winds']          ) 
                    HRW_data.t_corr_test      = np_append(HRW_data.t_corr_test     , channel[:]['t_corr_test']      )   
                    HRW_data.applied_QI       = np_append(HRW_data.applied_QI      , channel[:]['applied_QI']       )  
                    HRW_data.NWP_wind_levels  = np_append(HRW_data.NWP_wind_levels , channel[:]['NWP_wind_levels']  ) 
                    HRW_data.num_prev_winds   = np_append(HRW_data.num_prev_winds  , channel[:]['num_prev_winds']   )
                    HRW_data.orographic_index = np_append(HRW_data.orographic_index, channel[:]['orographic_index'] )
                    HRW_data.cloud_type       = np_append(HRW_data.cloud_type      , channel[:]['cloud_type']       )
                    HRW_data.wind_channel     = np_append(HRW_data.wind_channel    , channel[:]['wind_channel']     )
                    HRW_data.correlation      = np_append(HRW_data.correlation     , channel[:]['correlation']      )
                    HRW_data.pressure_error   = np_append(HRW_data.pressure_error  , channel[:]['pressure_error']   )

                # sort according to wind_id
                inds = HRW_data.wind_id.argsort()
                HRW_data.subset(inds) # changes HRW_data itself

                # sorting without conversion to numpy arrays 
                #[e for (wid,pwid) in sorted(zip(HRW_data.wind_id,HRW_data.prev_wind_id))]

    else:
        print "*** Error, no file found"
        print ""
        sat_ID = "no file"
        # but we continue the program in order to add an empty channel below 


    ## filter data according to the given optional arguments 
    #n1 = str(HRW_data.channel.size)
    #HRW_data = HRW_data.filter(**kwargs)   
    #print "    apply filters "+' ('+n1+'->'+str(HRW_data.channel.size)+')'

    chn_name="HRW"
    satscene[chn_name].HRW_basic    = HRW_basic.filter(**kwargs)     # returns new object (deepcopied and filtered)
    satscene[chn_name].HRW_detailed = HRW_detailed.filter(**kwargs)  # returns new object (deepcopied and filtered)
    satscene[chn_name].info['units'] = 'm/s'
    satscene[chn_name].info['satname'] = 'meteosat'
    satscene[chn_name].info['satnumber'] = sat_ID
    satscene[chn_name].info['instrument_name'] = 'seviri'
    satscene[chn_name].info['time'] = satscene.time_slot
    satscene[chn_name].info['is_calibrated'] = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2011 SMHI

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Minimal script for assembling and processing segmented polar data.

We take the case of level 1b data (calibrated and geolocalized) from noaa 19,
as output from AAPP.

- Install mpop and pyresample
- Don't forget to set up the PPP_CONFIG_DIR variable to point to your
  configuration files.
- Edit the gdsmetop02.cfg configuration file (a template is provided in case
  you don't have one) with your data directory:
  
  .. code-block:: ini
  
   [avhrr-granules]
   type=eps_avhrr
   granularity=60
   full_scan_period=0.1667
   scan_width=2048
   dir=/data/prod/satellit/ears/avhrr
   filename=AVHR_xxx_1B_M02_%Y%m%d%H%M*


- Here is a minimal script that monitors a directory and builds composites:
"""
import sys
from datetime import timedelta, datetime
import glob
import os
import time

from mpop.saturn.gatherer import Granule, Gatherer


def get_files_newer_than(directory, time_stamp):
    """Get the list of files from the *directory* which are newer than a given
 *time_stamp*.
    """
    filelist = glob.glob(os.path.join(directory, "*"))
    return [filename for filename in filelist
            if datetime.fromtimestamp(os.stat(filename)[8]) > time_stamp]


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: " + sys.argv[0] + " directory wait_for_more"
        sys.exit()
        
    directory = sys.argv[1]
    # if we wait for files in the directory forever or not
    wait_for_more = eval(sys.argv[2])

    areas = ["euro4", "scan2"]

    gatherer = None

    time_stamp = datetime(1970, 1, 1)
    
    while True:

        # Scanning directory
        
        new_time_stamp = datetime.now()
        filenames = get_files_newer_than(directory, time_stamp)
        time_stamp = new_time_stamp

        # Adding files to the gatherer
        
        for filename in filenames:
            granule = Granule(filename)
            if gatherer is None:
                gatherer = Gatherer(areas_of_interest=areas,
                                    timeliness=timedelta(minutes=150),
                                    satname=granule.satname,
                                    number=granule.number,
                                    variant=granule.variant)
            gatherer.add(granule)

        # Build finished swath and process them.
            
        for swath in gatherer.finished_swaths:
            global_data = swath.concatenate()

            local_data = global_data.project(swath.area)

            time_string = global_data.time_slot.strftime("%Y%m%d%H%M")

            area_id = swath.area.area_id
            
            img = local_data.image.overview()
            img.save("overview_" + area_id + "_" + time_string + ".png")

            img = local_data.image.natural()
            img.save("natural_" + area_id + "_" + time_string + ".png")

        if not wait_for_more:
            break
        
        # wait 60 seconds before restarting
        time.sleep(60)
